"""
    Test that the BMI version of Barrier 3D (Version 2.0) is equivalent to Ian's original code in Version 1.0. Note that
    the input files for Version 1.0 are located in `V1_NoBMI/Parameter` and for Version 2.0 in 'tests/test_params/'.
    Here we test the model with SHRUBS ON.

    NOTE to users:
      - if using Barrier3D for the first time, remember to $ pip install -e .
"""

import time

import matplotlib.pyplot as plt

from barrier3d import Barrier3dBmi
from barrier3d.tools import plot as B3Dfunc

# specify data directories with initial conditions
datadir_V1 = "V1_NoBMI/"
datadir_V2 = "tests/test_params/"

# Version 2.0 ------------------------------
# create an instance of the new BMI class, which is the model
barrier3d = Barrier3dBmi()
input_file = "barrier3d-default-parameters.yaml"
barrier3d.initialize(datadir_V2 + input_file)

# increase time step
Time = time.time()
for time_step in range(1, barrier3d._model._TMAX):
    barrier3d.update()
    print("\r", "Time Step: ", time_step, end="")
SimDuration = time.time() - Time
print()
print("Elapsed Time: ", SimDuration, "sec")  # Print elapsed time of simulation

# Version 1.0 ------------------------------

# starts running immediately and ends with plots!
execfile(datadir_V1 + "Barrier3D.py")

# ----------------------------- #

# Plot 1: Dune Height Over Time (input in decameter)
B3Dfunc.plot_dune_height(barrier3d._model._DuneDomain, barrier3d._model._Dmax)
B3Dfunc.plot_dune_height(DuneDomain, Dmax)

# check that shoreline change is the same between the two and that shoreface slope starts in equilibrium
plt.figure()
plt.plot(x_s_TS, "b")
plt.plot(barrier3d._model.x_s_TS, "g")

plt.figure()
plt.plot(s_sf_TS, "b")
plt.plot(barrier3d._model._s_sf_TS, "g")



#! /usr/bin/env python
"""Basic Model Interface implementation for River Module"""
import pathlib

import numpy as np
from bmipy import Bmi

from .barrier3d import Barrier3d


class Barrier3dBmi(Bmi):
    _name = "Barrier3D"
    _input_var_names = ()
    _output_var_names = (
        "overwash_flux",
        "shoreface_flux",
        "shoreface_toe_position",
        "shoreline_position",
        "back_barrier_shoreline_position",
        "height_of_barrier",
    )

    def __init__(self):
        """Create a Barrier3dBmi module that is ready for initialization."""
        self._model = None
        self._values = {}
        self._var_units = {}

    @staticmethod
    def _get_file_prefix(filepath):
        """Return the prefix for a barrier3d input file.

        The prefix of a barrier3d input file is the part before the last `-` of
        the file name. If the file name doens't contain a `-`, then the prefix
        is `None`.

        Examples
        --------
        >>> Barrier3dBmi._get_file_prefix("barrier3d-default-parameters.yaml")
        'barrier3d-default'
        >>> Barrier3dBmi._get_file_prefix("parameters.yaml") is None
        True
        """
        parts = pathlib.Path(filepath).stem.rsplit("-", maxsplit=1)
        if len(parts) == 1:
            prefix = None
        else:
            prefix = parts[0]
        return prefix

    def initialize(self, config_file):
        filepath = pathlib.Path(config_file)

        parameter_file_prefix = Barrier3dBmi._get_file_prefix(filepath)

        self._model = Barrier3d.from_yaml(filepath.parent, prefix=parameter_file_prefix)

        self._values = {
            "overwash_flux": lambda: np.array(self._model.QowTS[:]),
            "shoreface_flux": lambda: np.array(self._model.QsfTS[:]),
            "shoreface_toe_position": lambda: np.array(self._model.x_t_TS[:]),
            "shoreline_position": lambda: np.array(self._model.x_s_TS[:]),
            "back_barrier_shoreline_position": lambda: np.array(self._model.x_b_TS[:]),
            "height_of_barrier": lambda: np.array(self._model.h_b_TS[:]),
        }

        self._var_units = {
            "overwash_flux": "m^3 / m",
            "shoreface_flux": "m^3 / m",
            "shoreface_toe_position": "dam",
            "shoreline_position": "dam",
            "back_barrier_shoreline_position": "dam",
            "height_of_barrier": "dam",
        }

        self._var_type = {}
        for name in self._input_var_names + self._output_var_names:
            self._var_type[name] = str(np.dtype(float))

    def update(self):
        self._model.update()
        self._model.update_dune_domain()

    def update_frac(self, time_frac):
        """Update model by a fraction of a time step."""
        raise NotImplementedError("update_frac")

    def update_until(self, then):
        """Update model until a particular time."""
        raise NotImplementedError("update_until")

    def finalize(self):
        pass

    def get_var_type(self, name):
        return self._var_type[name]

    def get_var_units(self, name):
        return self._var_units[name]

    def get_var_nbytes(self, name):
        return self._values[name]().nbytes

    def get_var_itemsize(self, name):
        return np.dtype(self.get_var_type(name)).itemsize

    def get_var_size(self, name):
        return self._values[name]().size

    def get_var_location(self, name):
        return "none"

    def get_var_grid(self, name):
        raise NotImplementedError("get_var_grid")

    def get_grid_rank(self, grid):
        raise NotImplementedError("get_grid_rank")

    def get_grid_size(self, grid):
        raise NotImplementedError("get_grid_size")

    def get_grid_shape(self, grid, shape):
        raise NotImplementedError("get_grid_shape")

    def get_grid_spacing(self, grid, spacing):
        raise NotImplementedError("get_grid_spacing")

    def get_grid_origin(self, grid, origin):
        raise NotImplementedError("get_grid_origin")

    def get_grid_type(self, grid):
        raise NotImplementedError("get_grid_type")

    def get_value(self, name, dest):
        dest[:] = self._values[name]()
        return dest

    def set_value(self, var_name, new_vals):
        """Set model values."""
        raise NotImplementedError("set_value")

    def get_component_name(self):
        return self._name

    def get_input_var_names(self):
        return self._input_var_names

    def get_output_var_names(self):
        return self._output_var_names

    def get_input_item_count(self):
        return len(self._input_var_names)

    def get_output_item_count(self):
        return len(self._output_var_names)

    def get_start_time(self):
        return 0.0

    def get_end_time(self):
        return float(self._model._TMAX)

    def get_current_time(self):
        return float(self._model.time_step)

    def get_time_step(self):
        return 1.0

    def get_time_units(self):
        return "yr"

    def get_grid_node_count(self, grid):
        raise NotImplementedError("get_grid_node_count")

    def get_grid_edge_count(self, grid):
        raise NotImplementedError("get_grid_edge_count")

    def get_grid_face_count(self, grid):
        raise NotImplementedError("get_grid_face_count")

    def get_grid_edge_nodes(self, grid, edge_nodes):
        raise NotImplementedError("get_grid_edge_nodes")

    def get_grid_face_edges(self, grid, face_edges):
        raise NotImplementedError("get_grid_face_edges")

    def get_grid_face_nodes(self, grid, face_nodes):
        raise NotImplementedError("get_grid_edge_nodes")

    def get_grid_nodes_per_face(self, grid, nodes_per_face):
        raise NotImplementedError("get_grid_nodes_per_face")

    def get_grid_x(self, grid, x):
        raise NotImplementedError("get_grid_x")

    def get_grid_y(self, grid, y):
        raise NotImplementedError("get_grid_y")

    def get_grid_z(self, grid, z):
        raise NotImplementedError("get_grid_z")

    def get_value_at_indices(self, name, dest, inds):
        raise NotImplementedError("get_value_at_indices")

    def get_value_ptr(self, name):
        raise NotImplementedError("get_value_ptr")

    def set_value_at_indices(self, name, ids, src):
        raise NotImplementedError("set_value_at_indices")

    @property
    def model(self):
        return self._model



"""
    Example run script for Version 1.0 of Barrier3D (NoBMI). Note, you must be in the top Barrier3D working directory.
"""
import version1_local_copy.Barrier3D_Functions as B3Dfunc

datadir = "version1_local_copy/"

TMAX = []
t = []
DuneDomain = []
DomainTS = []
Dmax = []

# starts running immediately and ends with plots!
exec(open(datadir + "Barrier3D.py").read())

# Plot 1: Dune Height Over Time (input in decameter)
B3Dfunc.plot_DuneHeight(DuneDomain, Dmax)



import os
import pathlib
import textwrap
from contextlib import suppress

from pydantic import (
    BaseModel,
    Field,
    FilePath,
    NonNegativeFloat,
    NonPositiveFloat,
    PositiveInt,
    confloat,
    conint,
    conlist,
    validator,
)


class Barrier3dConfiguration(BaseModel):
    elevation_file: FilePath = Field(
        "barrier3d-default-elevations.npy",
        description="File that contains initial elevations in [m MHH]",
    )
    dune_file: FilePath = Field(
        "barrier3d-default-dunes.npy",
        description="File that contains initial dune height values [m]",
    )
    growth_param_file: FilePath = Field(
        "barrier3d-default-growthparam.npy",
        description="File that contains initial growth parameters",
    )
    storm_file: FilePath = Field(
        "barrier3d-default-storms.npy", description="File that contains storm data"
    )

    TMAX: PositiveInt = Field(
        150,
        description="Duration of simulation",
        unit="y",
    )
    StormStart: PositiveInt = Field(
        2,
        description=(
            "Year when storm can start occurring (NOTE: if changed, need new storm "
            "time series)"
        ),
        unit="y",
    )
    BarrierLength: NonNegativeFloat = Field(
        500.0,
        description="Static length (alongshore) of island segment",
        unit="m",
    )
    DuneWidth: NonNegativeFloat = Field(
        20.0,
        description=(
            "Width (cross-shore) of island dune field; for illustration purposes only"
        ),
        unit="m",
    )
    LShoreface: NonNegativeFloat = Field(
        500.0,
        description="Length of shoreface",
        unit="m",
    )
    DShoreface: NonNegativeFloat = Field(
        10.0,
        description="Height of shoreface",
        unit="m",
    )
    BayDepth: NonNegativeFloat = Field(
        3.0,
        description="Depth of bay behind island segment",
        unit="m",
    )
    MHW: NonNegativeFloat = Field(
        0.46,
        description=(
            "Elevation of Mean High Water (NOTE: if changed, need new storm time "
            "series)"
        ),
        unit="m NAVD88",
    )
    Dstart: NonNegativeFloat = Field(
        0.5, description="Initial height of dune domain above berm elevation", unit="m"
    )
    BermEl: NonNegativeFloat = Field(
        1.9,
        description=(
            "Static elevation of berm; berm elevation + dune height = dune elevation "
            "(NOTE: if changed, need new MSSM and storms)"
        ),
        unit="m NAVD88",
    )
    rmin: float = Field(
        0.35,
        description="Minimum growth rate for logistic dune growth",
    )
    rmax: float = Field(
        0.85,
        description="Maximum growth rate for logistic dune growth",
    )
    HdDiffu: NonNegativeFloat = Field(
        0.75,
        description=(
            "Dune diffusion parameter (i.e. max height offset between adjacent dune "
            "cells)"
        ),
        unit="m",
    )
    Dmaxel: NonNegativeFloat = Field(
        3.4,
        description="Maximum elevation of dunes",
        unit="m NAVD88",
    )
    C1: float = Field(8.8, description="Empirical dune erosion parameter", unit="m")
    C2: float = Field(4.6, description="Empirical dune erosion parameter", unit="m")
    DuneRestart: NonNegativeFloat = Field(
        0.075,
        description="Restart height for dunes lowered to essentially zero",
        unit="m",
    )
    Rat: float = Field(
        0.0,
        description=(
            "Rate of shoreline retreat attributed to alongshore transport; "
            "(-) = erosion, (+) = accretion"
        ),
        unit="m / y",
    )
    RSLR_Constant: bool = Field(
        True,
        description=(
            "Relative sea-level rise rate will be constant, otherwise logistic growth "
            "function used for time series"
        ),
    )
    RSLR_const: float = Field(
        0.004,
        description="Relative sea-level rise rate",
        unit="m / y",
    )
    beta: NonNegativeFloat = Field(
        0.04,
        description="Beach slope for runup calculations",
    )
    StormSeries: conlist(item_type=NonNegativeFloat) = Field(
        [],
        description="Time series of storms",
    )
    nn: float = Field(
        0.5,
        description="Flow routing constant",
    )
    mm: float = Field(
        2.0,
        description="Exponent constant for sediment transport",
    )
    Rin_r: float = Field(
        2.0,
        description=(
            "Run-up regime infiltration rate (volume of overwash flow lost per m "
            "cross-shore per time step)"
        ),
    )
    Rin_i: float = Field(
        0.25,
        description=(
            "Inundation regime infiltration rate (volume of overwash flow lost per m "
            "cross-shore per time step)"
        ),
    )
    Qs_min: float = Field(
        1.0,
        description="Minimum discharge needed for sediment transport",
        unit="m^3 / hr",
    )
    MaxUpSlope: float = Field(
        0.25,
        description="Maximum slope water can flow upward",
        unit="m / m",
    )
    threshold_in: float = Field(
        0.25,
        description="Threshold to determine if in inundation regime",
        unit="m^3 / hr",
    )
    Kr: float = Field(
        0.000075,
        description="Sediment flux constant, run-up regime",
    )

    Ki: float = Field(
        0.0000075,
        description="Sediment flux constant, inundation regime",
    )
    Cbb_r: confloat(gt=0.0, lt=1.0) = Field(
        0.5,
        description=(
            "Coefficient for exponential decay of sediment load entering back-barrier "
            "bay in run-up regime"
        ),
    )
    Cbb_i: confloat(gt=0.0, lt=1.0) = Field(
        0.8,
        description=(
            "Coefficient for exponential decay of sediment load entering back-barrier "
            "bay in inundation regime"
        ),
    )
    Qs_bb_min: NonNegativeFloat = Field(
        1.0,
        description=(
            "Minimum sediment flux in back-barrier bay (below which sediment won't "
            "flux)"
        ),
        unit="m^3 / hr",
    )
    Cx: NonNegativeFloat = Field(
        10.0,
        description=(
            "Multiplier with the average slope of the interior for constant C in "
            "inundation transport rule"
        ),
    )
    OWss_i: conint(ge=1) = Field(2, description="Overwash substep")
    OWss_r: conint(ge=1) = Field(1, description="Overwash substep")
    k_sf: float = Field(
        5000,
        description="Shoreface flux rate constant",
        unit="m^3 / m / y",
    )
    s_sf_eq: float = Field(
        0.02,
        description="Equilibrium shoreface slope",
    )
    Shrub_ON: bool = Field(
        False, description="1 = shrubs on in simulation, 0 = shrubs off"
    )
    Seedmin: NonNegativeFloat = Field(
        1000.0,
        description="Seeds produced per shrub per year (fecundity)",
        unit="1 / yr",
    )
    Seedmax: NonNegativeFloat = Field(
        10000.0,
        description="Seeds produced per shrub per year (fecundity)",
        unit="1 / yr",
    )
    disp_mu: float = Field(
        -0.721891,
        description="For lognormal probability distribution of seed dispersal distance",
    )
    disp_sigma: float = Field(
        1.5,
        description="For lognormal probability distribution of seed dispersal distance",
    )
    Dshrub: NonNegativeFloat = Field(
        2.75,
        description="Minimum elevation of fronting dune for shrub growth",
        unit="m",
    )
    GermRate: confloat(ge=0.0, le=1.0) = Field(0.6, description="Germination rate")
    TimeFruit: NonNegativeFloat = Field(
        5.0, description="Age shrubs need to be before they start fruiting", unit="y"
    )
    Female: confloat(ge=0.0, le=1.0) = Field(
        0.5, description="Percentage of shrubs that are female"
    )
    ShrubEl_min: NonNegativeFloat = Field(
        1.2,
        description="Elevation range for shrub growth, minimum bound",
        unit="m NAVD88",
    )
    ShrubEl_max: NonNegativeFloat = Field(
        2.3,
        description="Elevation range for shrub growth, maximum bound",
        unit="m NAVD88",
    )
    SprayDist: NonNegativeFloat = Field(
        170.0,
        description="Distance from ocean shoreline that shrubs can establish",
        unit="m",
    )
    BurialLimit: NonNegativeFloat = Field(
        0.75,
        description=(
            "Maximum percentage of height that a shrub can be buried up to before dying"
        ),
        unit="m",
    )
    UprootLimit: NonPositiveFloat = Field(
        -0.2, description="Shrubs eroded beyond this limit killed", unit="m"
    )
    SalineLimit: NonNegativeFloat = Field(
        0.05,
        description="Dishcharge limit to determine shrub mortality via saline flooding",
        unit="m^3 / hr",
    )
    Qshrub_max: confloat(ge=0.0, le=1.0) = Field(
        0.15,
        description=(
            "Maximum percentage of overwash reduction through a shrub cell with full "
            "percent cover"
        ),
    )
    DuneParamStart: bool = Field(
        True, description="Dune height will come from external file"
    )
    GrowthParamStart: bool = Field(
        True, description="Dune growth parameters will come from external file"
    )
    MaxShrubHeight: NonNegativeFloat = Field(
        5.2, description="Maximum shrub height", unit="m"
    )
    ShorefaceToe: float = Field(0.0, description="Start location of shoreface toe [m]")
    SeededRNG: bool = Field(
        True, description="Use seeded random number generator for reproducibility"
    )

    @validator("rmax", always=True)
    def check_rmax(cls, v, values, **kwargs):  # noqa: B902
        min_field = "rmin"

        with suppress(KeyError):
            if v < values[min_field]:
                raise ValueError(
                    f"ensure this value is greater than or equal to {min_field} "
                    f"({values[min_field]})"
                )
        return v

    @validator("Seedmax", always=True)
    def check_seed_max(cls, v, values, **kwargs):  # noqa: B902
        min_field = "Seedmin"

        with suppress(KeyError):
            if v < values[min_field]:
                raise ValueError(
                    f"ensure this value is greater than or equal to {min_field} "
                    f"({values[min_field]})"
                )
        return v

    @validator("ShrubEl_max", always=True)
    def check_shrub_elevation_max(cls, v, values, **kwargs):  # noqa: B902
        min_field = "ShrubEl_min"

        with suppress(KeyError):
            if v < values[min_field]:
                raise ValueError(
                    f"ensure this value is greater than or equal to {min_field} "
                    f"({values[min_field]})"
                )
        return v

    def to_yaml(self):
        lines = []
        for k, meta in sorted(self._meta().items()):
            unit = meta["unit"]
            value = meta["value"]
            desc = meta["description"]

            if desc:
                lines += textwrap.wrap(
                    desc, width=88, initial_indent="# ", subsequent_indent="# "
                )
            lines += [f"{k}: {value!r}" + (f"  # [{unit}]" if unit else "")]

        return os.linesep.join(lines)

    def to_py(self):
        lines = []
        for k, meta in sorted(self._meta().items()):
            unit = meta["unit"]
            value = meta["value"]
            desc = meta["description"]

            if desc:
                lines += textwrap.wrap(
                    desc, width=88, initial_indent="# ", subsequent_indent="# "
                )
            lines += [f"{k} = {value!r}" + (f"  # [{unit}]" if unit else "")]

        return os.linesep.join(lines)

    def _meta(self):
        schema = self.schema()

        meta = {}
        for k, v in sorted(self.dict().items()):
            if isinstance(v, pathlib.Path):
                v = str(v)
            meta[k] = {
                "value": v,
                "unit": schema["properties"][k].get("unit", None),
                "description": schema["properties"][k].get("description", None),
            }

        return meta

    @classmethod
    def from_yaml(cls, filepath):
        import yaml

        with open(filepath) as fp:
            config = yaml.safe_load(fp)

        return cls(**config)

    @classmethod
    def from_py(cls, filepath):
        from importlib.machinery import SourceFileLoader

        valid_keys = set(cls.schema()["properties"])

        mod = SourceFileLoader("config", str(filepath)).load_module()
        config = {k: v for k, v in mod.__dict__.items() if k in valid_keys}

        return cls(**config)



import pathlib

import numpy as np
import pytest
from packaging.version import Version

from barrier3d import Barrier3dBmi

# Version 1.0 (original version written by Ian Reeves) ------------------------------
DATA_DIR = pathlib.Path(__file__).absolute().parent.parent / "version1_local_copy"


def run_old_version():
    from importlib.machinery import SourceFileLoader

    mod = SourceFileLoader("v1", str(DATA_DIR / "Barrier3D.py")).load_module()

    return {
        k: mod.__dict__[k]
        for k in ["DuneDomain", "x_s_TS", "ShorelineChangeTS", "s_sf_TS", "QowTS"]
    }


@pytest.mark.skipif(
    Version(np.__version__) >= Version("1.24"),
    reason="numpy<1.24 required for barrier3d v1",
)
@pytest.mark.slow
def test_BMI_against_V1(tmpdir, datadir):
    """check that the BMI and Version 1 of Barrier3D (by Ian Reeves) are equivalent."""
    with tmpdir.as_cwd():
        expected = run_old_version()

    # Version 2.0 and beyond (contains a BMI) ------------------------------
    with tmpdir.as_cwd():
        barrier3d = Barrier3dBmi()
        barrier3d.initialize(datadir / "barrier3d-default-parameters.yaml")

        # increase time step
        for _ in range(1, barrier3d._model._TMAX):
            barrier3d.update()

    # dune height over time
    # assert np.all(barrier3d._model._DuneDomain[-1, :, 0] == DuneDomain[-1, :, 0])

    # neither of these worked
    # assert_array_almost_equal(
    #     barrier3d._model._DuneDomain[-1, :, 0], DuneDomain[-1, :, 0]
    # )

    # shoreline change time series, BMI consistently less
    assert np.all(barrier3d._model._x_s_TS == expected["x_s_TS"])

    # dune migration time series
    assert np.all(barrier3d._model._ShorelineChangeTS == expected["ShorelineChangeTS"])

    # shoreface slope time series
    assert np.all(barrier3d._model._s_sf_TS == expected["s_sf_TS"])

    # because we are producing less overwash
    assert np.all(barrier3d._model._QowTS == expected["QowTS"])



import matplotlib.pyplot as plt
import numpy as np


def plot_dune_height(dune_height, max_dune_height):
    dune_crest = dune_height.max(axis=2)

    fig = plt.figure(figsize=(14, 8))
    plt.rcParams.update({"font.size": 13})
    ax = fig.add_subplot(111)
    ax.matshow(
        (dune_crest) * 10,
        origin="lower",
        cmap="bwr",
        aspect="auto",
        vmin=0,
        vmax=max_dune_height * 10,
    )
    ax.xaxis.set_ticks_position("bottom")  # analysis:ignore

    plt.xlabel("Alongshore Distance (dam)")
    plt.ylabel("Year")
    plt.title("Dune Height (m)")

    fig.show()


def plot_shrub_percent_cover_tmax(PercentCoverTS, TMAX, DeadPercentCoverTS):
    ShrubPC = PercentCoverTS[TMAX - 1]
    shrubFig2 = plt.figure(figsize=(10, 5))
    ax = shrubFig2.add_subplot(111)
    cax = ax.matshow(
        ShrubPC, origin="lower", cmap="YlGn", vmin=0, vmax=1
    )  # analysis:ignore
    ax.xaxis.set_ticks_position("bottom")
    cbar = shrubFig2.colorbar(cax)
    cbar.set_label("Shrub Percent Cover", rotation=270, labelpad=20)

    Dead = DeadPercentCoverTS[TMAX - 1]
    Dy, Dx = np.argwhere(Dead > 0).T
    Dz = Dead[Dy, Dx] * 20
    ax.scatter(Dx, Dy, marker="x", s=Dz, color="k", alpha=0.25)

    plt.xlabel("Alongshore Distance (dm)")
    plt.ylabel("Cross-Shore Diatance (dm)")
    plt.title("Final Shrub Percent Cover")
    plt.show()

    shrubFig2.show()


def plot_ElevTMAX(
    TMAX,
    DuneDomain,
    DomainTS,
    BermEl,
    Shrub_ON,
    PercentCoverTS,
    DeadPercentCoverTS,
    DuneWidth,
):
    TMAX = TMAX - 1
    Dunes = (DuneDomain[TMAX, :, :] + BermEl) * 10
    Dunes = np.rot90(Dunes)
    Dunes = np.flipud(Dunes)
    Domain = DomainTS[TMAX] * 10
    Domain = np.vstack([Dunes, Domain])
    if Shrub_ON == 1:
        Shrubs = PercentCoverTS[TMAX]
        Dead = DeadPercentCoverTS[TMAX]
        Sy, Sx = np.argwhere(Shrubs > 0).T
        Sz = Shrubs[Sy, Sx] * 80
        Dy, Dx = np.argwhere(Dead > 0).T
        Dz = Dead[Dy, Dx] * 80
    elevFig1 = plt.figure(figsize=(14, 5))
    ax = elevFig1.add_subplot(111)
    ax.matshow(
        Domain, origin="lower", cmap="terrain", vmin=-1.1, vmax=4.0
    )  # , interpolation='gaussian') # analysis:ignore
    if Shrub_ON == 1:
        ax.scatter(
            Sx,
            Sy + DuneWidth,
            marker="$*$",
            s=Sz,
            c="black",
            alpha=0.7,
            edgecolors="none",
        )
        ax.scatter(
            Dx,
            Dy + DuneWidth,
            marker="$*$",
            s=Dz,
            c="red",
            alpha=0.7,
            edgecolors="none",
        )
    ax.xaxis.set_ticks_position("bottom")
    # cbar = elevFig1.colorbar(cax)
    # cbar.set_label('Elevation (m)', rotation=270)
    plt.xlabel("Alongshore Distance (dam)")
    plt.ylabel("Cross-Shore Distance (dam)")
    plt.title("Interior Elevation (m)")
    timestr = "Time = " + str(TMAX) + " yrs"
    plt.text(1, 1, timestr)
    plt.tight_layout()
    plt.show()
    # name = "Output/FinalElevation"
    # elevFig1.savefig(name)
    plt.show()



import contextlib
import math
import os
import pathlib

import numpy as np
import pandas

from .configuration import Barrier3dConfiguration


@contextlib.contextmanager
def as_cwd(path):
    prev_cwd = os.getcwd()
    os.chdir(path)
    yield
    os.chdir(prev_cwd)


def load_inputs(path_to_folder, prefix="barrier3d-default", fmt="yaml"):
    """Load input parameter and elevation data for Barrier3d.

    Parameters
    ----------
    path_to_folder : str, or Path
        Path to folder that contains barrier3d input files.
    prefix : str, optional
        Filename prefix for input files.
    fmt : {'py', 'yaml'}, optional
        Input file format.

    Returns
    -------
    dict
        Processes input parameters for a barrier3d simulation.
    """
    prefix = f"{prefix}-" if prefix else ""

    fmts = ("py", "yaml")
    if fmt not in fmts:
        raise ValueError(
            "unrecognized format for input files ({fmt} not in {fmts})".format(
                fmt=fmt, fmts=", ".join(fmts)
            )
        )
    parameter_file = pathlib.Path(f"{prefix}parameters.{fmt}")

    with as_cwd(path_to_folder):
        if not parameter_file.is_file():
            raise FileNotFoundError(parameter_file.absolute())

        params = load_parameters(parameter_file, fmt=fmt)

        params["InteriorDomain"] = load_elevation(
            params.pop("elevation_file"), fmt=None
        )  # interior elevations must come from an external file

        params["StormSeries"] = load_storms(
            params.pop("storm_file"), fmt=None
        )  # storms must come from a time series

        if params[
            "DuneParamStart"
        ]:  # user specifies if dune height will come from external file
            params["DuneStart"] = load_dunes(params.pop("dune_file"), fmt=None)
        else:
            params.pop("dune_file")  # get rid of the keyword

        if params[
            "GrowthParamStart"
        ]:  # user specifies if growth parameters will come from external file
            params["GrowthStart"] = load_growth_param(
                params.pop("growth_param_file"), fmt=None
            )
        else:
            params.pop("growth_param_file")

        _process_raw_input(params)

    return params


def load_parameters(path_to_file, fmt="yaml"):
    try:
        loader = getattr(Barrier3dConfiguration, f"from_{fmt}")
    except AttributeError as error:
        raise ValueError(f"format not understood ({fmt})") from error
    else:
        return loader(path_to_file).dict()


def load_elevation(path_to_file, fmt="npy"):
    fmt = fmt or _guess_format(path_to_file)

    # load_elevation(path_to_file, fmt="npy"):
    if fmt == "npy":
        elevation = np.load(path_to_file, allow_pickle=True)
    elif fmt == "csv":
        elevation = np.loadtxt(path_to_file, delimiter=",")
        # data = pandas.read_csv(
        #     path_to_csv,
        #     names=("x", "y", "z"),
        #     dtype={"x": int, "y": int, "z": float},
        #     comment="#",
        #     header=0,
        # )
        # elevation = np.zeros((data.x.max() + 1, data.y.max() + 1), dtype=float)
        # elevation[(data.x, data.y)] = data.z
    else:
        fmts = ", ".join(["npy", "csv"])
        raise ValueError(
            f"unrecognized format for storm time series ({fmt} not one of {fmts})"
        )

    return elevation


def _guess_format(path_to_file):
    """Guess the format of a file based on its name.

    Parameters
    ----------
    path_to_file : str, path-like
        Name of the file.

    Returns
    -------
    str
        The format of the file or an empty string if it can't be determined.

    Examples
    --------
    >>> from barrier3d.load_input import _guess_format
    >>> _guess_format("file.csv")
    'csv'
    >>> _guess_format("file.npy")
    'npy'
    >>> _guess_format("file-with-an-extension")
    ''
    """
    return pathlib.Path(path_to_file).suffix[1:]


def load_storms(path_to_file, fmt="npy"):
    fmt = fmt or _guess_format(path_to_file)

    if fmt == "npy":
        data = np.load(path_to_file)
    elif fmt == "csv":
        df = pandas.read_csv(
            path_to_file,
            names=("time", "Rhigh", "Rlow", "period", "duration"),
            dtype={
                "time": int,
                "Rhigh": float,
                "Rlow": float,
                "period": float,
                "duration": int,
            },
            comment="#",
            header=0,
        )
        df.to_numpy()
        data = np.hstack(
            (
                df["time"][:, None],
                df["Rhigh"][:, None],
                df["Rlow"][:, None],
                df["period"][:, None],
                df["duration"][:, None],
            )
        )
    else:
        fmts = ", ".join(["npy", "csv"])
        raise ValueError(
            f"unrecognized format for storm time series ({fmt} not one of {fmts})"
        )

    return data


def load_dunes(path_to_file, fmt="npy"):
    fmt = fmt or _guess_format(path_to_file)

    if fmt == "npy":
        data = np.load(path_to_file)
    elif fmt == "csv":
        data = pandas.read_csv(
            path_to_file, names=("DuneStart",), dtype=float, comment="#", header=0
        )["DuneStart"].values
    else:
        fmts = ", ".join(["npy", "csv"])
        raise ValueError(f"unrecognized format for dunes ({fmt} not one of {fmts})")

    return data


def load_growth_param(path_to_file, fmt="npy"):
    fmt = fmt or _guess_format(path_to_file)

    if fmt == "npy":
        data = np.load(path_to_file)
    elif fmt == "csv":
        data = pandas.read_csv(
            path_to_file, names=("growth_param",), dtype=float, comment="#", header=0
        )["growth_param"].values
    else:
        fmts = ", ".join(["npy", "csv"])
        raise ValueError(
            f"unrecognized format for growth param ({fmt} not one of {fmts})"
        )

    return data


def _process_raw_input(params):
    params["RNG"] = np.random.default_rng(seed=1973)

    params["TMAX"] = int(params["TMAX"])
    params["StormStart"] = int(params["StormStart"])

    params.setdefault("StormSeries", [])

    params["LShoreface"] /= 10.0
    params["DShoreface"] /= 10.0
    params["BayDepth"] /= 10.0
    params["MHW"] /= 10.0
    params["BarrierLength"] = int(params["BarrierLength"] / 10.0)
    params["DuneWidth"] = int(params["DuneWidth"] / 10.0)

    if params["InteriorDomain"].shape[1] > params["BarrierLength"]:
        params["InteriorDomain"] = params["InteriorDomain"][
            :, : params["BarrierLength"]
        ]
    else:
        params["BarrierLength"] = params["InteriorDomain"].shape[1]

    params["DomainWidth"] = len(params["InteriorDomain"])

    params["Dstart"] /= 10.0
    params["BermEl"] = params["BermEl"] / 10.0 - params["MHW"]

    # Initialize dune crest height domain
    params["DuneDomain"] = np.zeros(
        (params["TMAX"], params["BarrierLength"], params["DuneWidth"])
    )

    if "DuneStart" in params:
        params["DuneDomain"][0, :, 0] = params["DuneStart"][0 : params["BarrierLength"]]
        params.pop("DuneStart")
    else:
        if params["SeededRNG"]:
            params["DuneDomain"][0, :, 0] = np.ones([1, params["BarrierLength"]]) * (
                params["Dstart"]
                + (
                    -0.01
                    + (0.01 - (-0.01))
                    * params["RNG"].random((1, params["BarrierLength"]))
                )
            )
        else:
            params["DuneDomain"][0, :, 0] = np.ones([1, params["BarrierLength"]]) * (
                params["Dstart"]
                + (
                    -0.01
                    + (
                        -0.01
                        + (0.01 - (-0.01)) * np.random.rand(1, params["BarrierLength"])
                    )
                )
            )
    params["DuneDomain"][0, :, 1:] = params["DuneDomain"][0, :, 0, None]

    if "GrowthStart" in params:
        params["growthparam"] = params["GrowthStart"][0 : params["BarrierLength"]]
        params.pop("GrowthStart")
    else:
        if params["SeededRNG"]:
            params["growthparam"] = params["rmin"] + (
                params["rmax"] - params["rmin"]
            ) * params["RNG"].random((1, params["BarrierLength"]))
        else:
            params["growthparam"] = params["rmin"] + (
                params["rmax"] - params["rmin"]
            ) * np.random.rand(1, params["BarrierLength"])

    params["HdDiffu"] /= 10.0

    # params.pop("rmin")
    # params.pop("rmax")

    # Maximum dune height
    # params["Dmaxel"] /= 10.0
    params["Dmaxel"] = params["Dmaxel"] / 10.0 - params["MHW"]

    # Erosion parameters
    params["DuneRestart"] /= 10.0

    # Volume of sediment lost via alongshore transport
    # params["Rat"] = params["Rat"] / 10.0
    params["Rat"] = params["Rat"] / -10.0
    params["Qat"] = params["Rat"] * params["DShoreface"]
    params.pop("Rat")

    # Relative Sea Level Rise Rate
    # params["RSLR"] /= 10.0
    if params["RSLR_Constant"]:
        # Constant RSLR
        params["RSLR_const"] /= 10.0
        params["RSLR"] = [params["RSLR_const"]] * params["TMAX"]
    else:
        # Logistic RSLR rate projection - Rohling et al. (2013): gives a logistic
        # growth curve starting from 0.003 m/yr (AD2000); note Figure 3a is in m/cyr
        # (divide by 100 to get m/yr)
        params["RSLR"] = []
        # maximum achievable rate, m/yr -- probability maximum = 0.0075 m/y,
        # 68% upper bound = 0.022 m/yr
        alpha = 0.022
        # 0.003 is constant: corresponds to 0.003 m/yr at AD2000
        beta = alpha / 0.003 - 1
        # yr -- probability maximum = 450, 68% upper bound = 600 (KA: changed from
        # Ian's estimates)
        gamma = 600
        C = 12  # constant
        # this is range shown in Figure 3a of Rohling, starting at AD2000
        for t in range(0, params["TMAX"]):
            delta = alpha / (1 + beta * math.exp(-t / gamma * C))
            params["RSLR"].append(delta / 10)  # Convert from m/yr to dam/yr

    # Shoreface
    params["x_t"] = params["ShorefaceToe"] / 10.0
    params.pop("ShorefaceToe")

    # Shrubs
    params["Dshrub"] /= 10.0
    params["ShrubEl_min"] = params["ShrubEl_min"] / 10.0 - params["MHW"]
    params["ShrubEl_max"] = params["ShrubEl_max"] / 10.0 - params["MHW"]
    # params["TideAmp"] /= 10.0
    params["SprayDist"] /= 10.0
    params["MaxShrubHeight"] /= 10.0

    # # Shrub height (with age as proxy)
    # SH = np.linspace(0, 0.3, 10)
    # addend = np.ones(params["TMAX"] + 50)  # (years 10+)
    # params["SH"] = np.append(SH, addend)

    # Overwash Interaction (convert to decameters)
    params["BurialLimit"] /= 10.0
    params["UprootLimit"] /= 10.0
    params["Qs_min"] /= 1000
    params["Qs_bb_min"] /= 1000

    # Percent cover change (years 0-9)
    PC = np.array([0, 0.04, 0.08, 0.10, 0.15, 0.15, 0.20, 0.35, 0.80, 1])
    addend = np.ones(params["TMAX"] + 50)  # (years 10+)
    params["PC"] = np.append(PC, addend)

    return params



import copy
import math
import random

import numpy as np

from .load_input import load_inputs


class Barrier3dError(Exception):
    pass


class Barrier3d:
    def SeaLevel(self, InteriorDomain, DuneDomain, t):
        """Accounts for relative sea level rise, returns updated elevation domains."""

        # Decrease all elevation this year by RSLR increment
        InteriorDomain = InteriorDomain - self._RSLR[t]
        DuneDomain[t - 1] = DuneDomain[t - 1] - self._RSLR[t]

        # Bay can't be deeper than BayDepth (roughly equivalent to constant
        # back-barrier slope)
        InteriorDomain[InteriorDomain < -self._BayDepth] = -self._BayDepth

        return InteriorDomain, DuneDomain

    def FindWidths(self, InteriorDomain, SL):
        """Finds DomainWidth and InteriorWidth
        DomainWidth is wide enough to fit widest part of island; InteriorWidth
        stores the width of the island at each row of the domain where elevation
        is greater than 0
        """

        DomainWidth = np.shape(InteriorDomain)[0]  # (dam) analysis:ignore
        InteriorWidth = [0] * self._BarrierLength
        for bl in range(self._BarrierLength):
            width = next(
                (
                    index
                    for index, value in enumerate(InteriorDomain[:, bl])
                    if value <= SL
                ),
                DomainWidth,
            )
            width = width - 1
            if width < 0:
                width = 0
            InteriorWidth[bl] = width
        # Average width of island
        InteriorWidth_Avg = np.nanmean(InteriorWidth)

        return DomainWidth, InteriorWidth, InteriorWidth_Avg

    def DuneGrowth(self, DuneDomain, t):
        """Grows dune cells"""

        # Set max dune height
        Dmax = self._Dmaxel - self._BermEl  # (dam) Max dune height
        if Dmax < 0:
            Dmax = 0

        Cf = 3  # Decay coefficient
        Qdg = 0
        # Grow dune
        for q in range(self._DuneWidth):
            reduc = 1 / (Cf**q)
            G = (
                self._growthparam
                * DuneDomain[t - 1, :, q]
                * (1 - DuneDomain[t - 1, :, q] / Dmax)
                * reduc
            )
            DuneDomain[t, :, q] = G + DuneDomain[t - 1, :, q]
            Qdg = Qdg + (
                np.sum(G) / self._BarrierLength
            )  # Volume of sediment lost from beach/shoreface from dune growth

        return DuneDomain, Dmax, Qdg

    def Shrubs(
        self,
        InteriorDomain,
        DuneDomainCrest,
        ShrubDomainFemale,
        ShrubDomainMale,
        ShrubPercentCover,
        DeadPercentCover,
        BurialDomain,
        InteriorWidth_Avg,
        DomainWidth,
        ShrubDomainDead,
    ):
        """Main shrub expansion and mortality function"""

        ShrubDomainAll = ShrubDomainFemale + ShrubDomainMale

        BeachWidth = int(self._BermEl / self._beta)

        # Burial / Uprooting
        ShrubHeight = ShrubPercentCover * self._MaxShrubHeight
        DeadHeight = DeadPercentCover * self._MaxShrubHeight

        for col in range(len(ShrubPercentCover[0])):
            for w in range(len(ShrubPercentCover)):
                if ShrubDomainAll[w, col] > 0:
                    if (
                        BurialDomain[w, col] > ShrubHeight[w, col]
                        or BurialDomain[w, col] < self._UprootLimit
                    ):
                        ShrubDomainFemale[w, col] = 0
                        ShrubDomainMale[w, col] = 0

                    elif BurialDomain[w, col] > (
                        self._BurialLimit * ShrubHeight[w, col]
                    ):
                        ShrubDomainDead[w, col] = ShrubDomainAll[w, col]
                        ShrubDomainFemale[w, col] = 0
                        ShrubDomainMale[w, col] = 0

                elif ShrubDomainDead[w, col] > 0:
                    if (
                        BurialDomain[w, col] > DeadHeight[w, col]
                        or BurialDomain[w, col] < self._UprootLimit
                    ):
                        ShrubDomainDead[w, col] = 0

        tempAll = ShrubDomainFemale + ShrubDomainMale + ShrubDomainDead
        BurialDomain[tempAll == 0] = 0  # Reset burial domain

        ShrubDomainAll = ShrubDomainFemale + ShrubDomainMale  # Recalculate

        # Inundation
        # Kill shrubs that have fallen below minimum elevation, remove shrubs that
        # have fallen below Mean High Water (i.e. elevation of 0) (passive loss from
        # rising back-barrier water elevations)
        for w in range(DomainWidth):
            for col in range(self._BarrierLength):
                if InteriorDomain[w, col] < self._ShrubEl_min:
                    if InteriorDomain[w, col] > self._SL:
                        if ShrubDomainFemale[w, col] > 0 or ShrubDomainMale[w, col] > 0:
                            ShrubDomainDead[w, col] = (
                                ShrubDomainFemale[w, col] + ShrubDomainMale[w, col]
                            )  # Shrub remains as dead if above MHW
                            ShrubDomainFemale[w, col] = 0
                            ShrubDomainMale[w, col] = 0
                    else:
                        ShrubDomainDead[w, col] = 0  # Remove dead shrubs below MHW
                        ShrubDomainFemale[w, col] = 0
                        ShrubDomainMale[w, col] = 0

        ShrubDomainAll = ShrubDomainFemale + ShrubDomainMale  # Recalculate

        # Age the shrubs in years
        ShrubDomainFemale[ShrubDomainFemale > 0] += 1
        ShrubDomainMale[ShrubDomainMale > 0] += 1

        # ================================================
        # Drop Seed
        # Randomly drop a seed onto the island each time step
        if self._seeded_RNG_on:
            randX = self._RNG.integers(0, self._BarrierLength)
            randY = self._RNG.integers(0, max(1, InteriorWidth_Avg))
        else:
            randX = np.random.randint(0, self._BarrierLength)
            randY = np.random.randint(0, max(1, InteriorWidth_Avg))
        if (
            ShrubDomainFemale[randY, randX] == 0
            and ShrubDomainMale[randY, randX] == 0
            and ShrubDomainDead[randY, randX] == 0
            and DuneDomainCrest[randX] + self._BermEl >= self._Dshrub
            and InteriorDomain[randY, randX] >= self._ShrubEl_min
            and InteriorDomain[randY, randX] <= self._ShrubEl_max
        ):
            if self._seeded_RNG_on:
                if self._RNG.uniform() > self._Female:
                    ShrubDomainFemale[randY, randX] = 1
                else:
                    ShrubDomainMale[randY, randX] = 1
            else:
                if np.random.uniform() > self._Female:
                    ShrubDomainFemale[randY, randX] = 1
                else:
                    ShrubDomainMale[randY, randX] = 1

        # # Randomly drop a seed onto the LEFT SIDE of the island each time step
        # # Can drop anywhere with the first 2% of interior columns
        # # randX = np.random.randint(0,BarrierLength*0.02)
        # # randY = np.random.randint(0,max(1,InteriorWidth_Avg))
        # randX = self._RNG.integers(0,self.BarrierLength*0.02)
        # randY = self._RNG.integers(0,max(1,InteriorWidth_Avg))
        # if (
        #     ShrubDomainFemale[randY,randX] == 0
        #     and ShrubDomainMale[randY,randX] == 0
        #     and ShrubDomainDead[randY,randX] == 0
        #     and DuneDomainCrest[randX] + self.BermEl >= self.Dshrub
        #     and InteriorDomain[randY,randX] >= self.ShrubEl_min
        #     and InteriorDomain[randY,randX] <= self.ShrubEl_max
        # ):
        #     if self._RNG.uniform() > self._Female:
        #         ShrubDomainFemale[randY, randX] = 1
        #     else:
        #         ShrubDomainMale[randY, randX] = 1
        # ================================================

        # ### Disperse seeds
        # Loop through each row of island width (i.e. from ocean to mainland side
        # of island)
        for k in range(self._BarrierLength):
            if 0 in ShrubDomainAll:
                # For all cells with a shrub
                FemaleShrubs = ShrubDomainFemale[:, k]
                fruiting_shrub = [
                    index
                    for index, value in enumerate(FemaleShrubs)
                    if value >= self._TimeFruit
                ]
                numShrubCells = len(fruiting_shrub)
                # Determine how many seeds in each cell
                if self._seeded_RNG_on:
                    Seedspercell = self._RNG.integers(
                        self._Seedmin, high=self._Seedmax, size=numShrubCells
                    )
                    # For each shrub cell, determine survival rate for the cell in
                    # this year
                    SeedSurvYear = self._GermRate * self._RNG.random(numShrubCells)
                else:
                    Seedspercell = np.random.randint(
                        self._Seedmin, high=self._Seedmax, size=numShrubCells
                    )
                    SeedSurvYear = self._GermRate * np.random.rand(numShrubCells)

                for i in range(numShrubCells):
                    # For each shrub cell producing seeds, generate a random # of
                    # random numbers each rep. a single seed
                    if self._seeded_RNG_on:
                        randseeds = self._RNG.random(Seedspercell[i])
                    else:
                        randseeds = np.random.rand(Seedspercell[i])
                    # Find how many seeds produced in each cell actually survive
                    Survivors = len(randseeds[randseeds < SeedSurvYear[i]])

                    # Determine distance, rounding to nearest integer
                    if Survivors > 0:
                        if self._seeded_RNG_on:
                            DispDist = np.round(
                                self._RNG.lognormal(
                                    self._disp_mu, self._disp_sigma, Survivors
                                )
                            )
                        else:
                            DispDist = np.round(
                                np.random.lognormal(
                                    self._disp_mu, self._disp_sigma, Survivors
                                )
                            )
                    else:
                        DispDist = []

                    # If there are seeds to disperse
                    if len(DispDist) > 0:
                        # Loop through each individual seed to disperse
                        for j in range(len(DispDist + 1)):
                            # Create a meshgrid to find coordinates of all points
                            # that are dropdistance from origin
                            if DispDist[j] > 0:
                                gridsize = int(DispDist[j] * 2 + 2)
                                X, Y = np.meshgrid(
                                    range(1, gridsize), range(1, gridsize)
                                )
                                # Sets coordinates of plant where seed is coming from
                                originX = fruiting_shrub[i]
                                originY = k
                                matOriginX = math.ceil(len(X) / 2)
                                matOriginY = matOriginX
                                # Find the distance from origin to every other point
                                # on island
                                distMat = np.round(
                                    np.sqrt(
                                        (X - matOriginX) ** 2 + (Y - matOriginY) ** 2
                                    )
                                )
                                # Find coordinates of each point on island that is
                                # drop distance away from origin
                                coords = np.where(distMat == DispDist[j])
                                row = coords[0]
                                col = coords[1]

                                # Randomly select one of those points as the target
                                # point - this means equal probability of dispersal
                                # in every direction (valid assumption for avian
                                # dispersal)
                                if len(col) == 0:
                                    targetY = originX
                                    targetX = originY
                                else:
                                    if self._seeded_RNG_on:
                                        xx = self._RNG.integers(0, (len(col)) - 1)
                                    else:
                                        xx = random.randint(0, (len(col)) - 1)
                                    matTargetX = col[xx]
                                    matTargetY = row[xx]

                                    targetY = originX + (matTargetX - matOriginX)
                                    targetX = originY + (matTargetY - matOriginY)

                                # Put a plant in the ground if
                                #   -the dropdistance>0,
                                #   -the target drop location is within the island
                                #    domain,
                                #   -there is no shrub at the receiving cell (male,
                                #    female, or dead),
                                #   -the receiving cell has a tall enough fronting dune
                                #   -the receiving cell is within elevation range
                                if (
                                    not (
                                        not (0 <= targetY < DomainWidth)
                                        or not (0 <= targetX < self._BarrierLength)
                                        or not (
                                            ShrubDomainFemale[targetY, targetX] == 0
                                        )
                                    )
                                    and ShrubDomainMale[targetY, targetX] == 0
                                    and self._ShrubEl_min
                                    <= InteriorDomain[targetY, targetX]
                                    <= self._ShrubEl_max
                                    and ShrubDomainDead[targetY, targetX] < 1
                                ):
                                    # Shrubs can establish if height of fronting dune
                                    # greater than threshold
                                    if (
                                        DuneDomainCrest[targetX] + self._BermEl
                                        >= self._Dshrub
                                    ):
                                        # Decide if the tree wll be a female or male
                                        if self._seeded_RNG_on:
                                            if self._RNG.uniform() > self._Female:
                                                ShrubDomainFemale[targetY, targetX] = 1
                                            else:
                                                ShrubDomainMale[targetY, targetX] = 1
                                        else:
                                            if np.random.uniform() > self._Female:
                                                ShrubDomainFemale[targetY, targetX] = 1
                                            else:
                                                ShrubDomainMale[targetY, targetX] = 1
                                    # Shrubs can establish without dune if located a
                                    # certain distance from shoreline...
                                    elif (
                                        BeachWidth + self._DuneWidth + targetY
                                        > self._SprayDist
                                    ):
                                        if self._seeded_RNG_on:
                                            if self._RNG.uniform() > 0.5:
                                                # ... but have about 50% chance of
                                                # survival. Decide if the tree wll be a
                                                # female or male
                                                if self._RNG.uniform() > self._Female:
                                                    ShrubDomainFemale[
                                                        targetY, targetX
                                                    ] = 1
                                                else:
                                                    ShrubDomainMale[
                                                        targetY, targetX
                                                    ] = 1
                                        else:
                                            if np.random.uniform() > 0.5:
                                                # ... but have about 50% chance of
                                                # survival. Decide if the tree wll be a
                                                # female or male
                                                if np.random.uniform() > self._Female:
                                                    ShrubDomainFemale[
                                                        targetY, targetX
                                                    ] = 1
                                                else:
                                                    ShrubDomainMale[
                                                        targetY, targetX
                                                    ] = 1

        ShrubDomainAll = ShrubDomainFemale + ShrubDomainMale

        return (
            ShrubDomainAll,
            ShrubDomainFemale,
            ShrubDomainMale,
            BurialDomain,
            ShrubDomainDead,
        )

    def DuneGaps(self, DuneDomain, Dow, bermel, Rhigh):
        """Returns tuple of [gap start index, stop index, avg Rexcess of each gap,
        alpha: ratio of TWL / dune height]"""
        gaps = []
        start = 0
        #        stop = 0
        i = start
        while i < (len(Dow) - 1):
            adjacent = Dow[i + 1] - Dow[i]
            if adjacent == 1:
                i = i + 1
            else:
                stop = i

                x = DuneDomain[Dow[start] : (Dow[stop] + 1)]
                Hmean = sum(x) / float(len(x))
                Rexcess = Rhigh - (Hmean + bermel)
                alpha = Rhigh / (Hmean + bermel)
                gaps.append([Dow[start], Dow[stop], Rexcess, alpha])

                start = stop + 1
                i = start
        if i > 0:
            stop = i - 1

            x = DuneDomain[Dow[start] : (Dow[stop] + 1)]
            if len(x) > 0:
                Hmean = sum(x) / float(len(x))
                Rexcess = Rhigh - (Hmean + bermel)
                alpha = Rhigh / (Hmean + bermel)
                gaps.append([Dow[start], Dow[stop], Rexcess, alpha])
        return gaps

    def DiffuseDunes(self, DuneDomain, t):
        """Dune height diffusion in alongshore direction: smoothes dune height by
        redistributing sand from high dune to neighboring low dune(s)"""

        for w in range(self._DuneWidth):
            # Alongshore direction
            for d in range(2, self._BarrierLength):  # Loop L to R
                # Calculate height difference
                Ldiff = DuneDomain[t, d, w] - DuneDomain[t, d - 1, w]
                # Subtract sand from taller dune cell and add to adjacent shorter one
                # (assumes sand in dunes is consrvd)
                if Ldiff > self._HdDiffu:
                    sub = (
                        Ldiff - self._HdDiffu
                    ) / 2  # Height difference in excess of maximum, divided by two
                    DuneDomain[t, d, w] = (
                        DuneDomain[t, d, w] - sub
                    )  # Lower dune that's too tall
                    DuneDomain[t, d - 1, w] = (
                        DuneDomain[t, d - 1, w] + sub
                    )  # Raise dune that's too short
            for d in range((self._BarrierLength - 2), 0, -1):  # Loop R to L
                Rdiff = DuneDomain[t, d, w] - DuneDomain[t, d + 1, w]
                if Rdiff > self._HdDiffu:
                    sub = (Rdiff - self._HdDiffu) / 2
                    DuneDomain[t, d, w] = DuneDomain[t, d, w] - sub
                    DuneDomain[t, d + 1, w] = DuneDomain[t, d + 1, w] + sub

        return DuneDomain

    def SalineFlooding(
        self,
        ShrubDomainWidth,
        ShrubDomainAll,
        ShrubDomainFemale,
        ShrubDomainMale,
        ShrubDomainDead,
        d,
        i,
        Q0,
    ):
        """Kill all immature (< 1 yr-old) shrubs that have been flooded
        beyond a threshold discharge (Tolliver et al., 1997)"""

        if d < (ShrubDomainWidth - 1) and i < (self._BarrierLength - 1):
            if Q0 >= self._SalineLimit and ShrubDomainAll[d, i] == 1:
                ShrubDomainDead[d, i] = (
                    ShrubDomainMale[d, i] + ShrubDomainFemale[d, i]
                )  # Transfer to dead shrub domain
                ShrubDomainFemale[d, i] = 0
                ShrubDomainMale[d, i] = 0

        return ShrubDomainFemale, ShrubDomainMale, ShrubDomainDead

    def UpdateBurial(
        self, BurialDomain, ElevationChange, ShrubDomainWidth, ShrubDomainAll
    ):
        """Updates amount of burial/erosion for each shrub"""

        BurialDomain = BurialDomain + ElevationChange[1 : ShrubDomainWidth + 1, :]
        BurialDomain[ShrubDomainAll == 0] = 0

        return BurialDomain

    def UpdateShrubDomains(
        self,
        DomainWidth,
        ShrubDomainWidth,
        ShrubDomainFemale,
        ShrubDomainMale,
        ShrubDomainAll,
        ShrubPercentCover,
        BurialDomain,
        ShrubDomainDead,
        DeadPercentCover,
    ):
        """Updates size of shrub domains"""

        if DomainWidth > ShrubDomainWidth:
            AddRows = np.zeros([DomainWidth - ShrubDomainWidth, self._BarrierLength])
            ShrubDomainFemale = np.vstack([ShrubDomainFemale, AddRows])
            ShrubDomainMale = np.vstack([ShrubDomainMale, AddRows])
            ShrubDomainDead = np.vstack([ShrubDomainDead, AddRows])
            ShrubDomainAll = ShrubDomainFemale + ShrubDomainMale
            ShrubPercentCover = np.vstack([ShrubPercentCover, AddRows])
            DeadPercentCover = np.vstack([DeadPercentCover, AddRows])
            BurialDomain = np.vstack([BurialDomain, AddRows])
        elif DomainWidth < ShrubDomainWidth:
            RemoveRows = ShrubDomainWidth - DomainWidth
            ShrubDomainFemale = ShrubDomainFemale[0:-RemoveRows, :]
            ShrubDomainMale = ShrubDomainMale[0:-RemoveRows, :]
            ShrubDomainDead = ShrubDomainDead[0:-RemoveRows, :]
            ShrubDomainAll = ShrubDomainFemale + ShrubDomainMale
            ShrubPercentCover = ShrubPercentCover[0:-RemoveRows, :]
            DeadPercentCover = DeadPercentCover[0:-RemoveRows, :]
            BurialDomain = BurialDomain[0:-RemoveRows, :]

        return (
            ShrubDomainFemale,
            ShrubDomainMale,
            ShrubDomainAll,
            ShrubPercentCover,
            BurialDomain,
            ShrubDomainDead,
            DeadPercentCover,
        )

    def InteriorNoise(self, InteriorDomain, InteriorWidth):
        """Adds elevation noise to flat parts of the interior to represent aeolian
        re-working during interstorm periods"""

        zthresh = 0.005  # dam
        ythresh = 2  # dam
        gap = 1  # dam

        # Loop through each column and
        for x in range(self._BarrierLength):
            col = InteriorDomain[0 : InteriorWidth[x], x]

            # find continuous sections that are too flat
            der = np.diff(col)
            der_under = np.where(der < zthresh)[0]

            if len(der_under) > 0:
                gaps = np.diff(der_under) > gap
                flat_start = np.insert(der_under[1:][gaps], 0, der_under[0])
                flat_stop = np.append(der_under[:-1][gaps], der_under[-1])

                # Add noise to flat sections
                for n in range(1, len(flat_stop)):
                    flat_length = flat_stop[n] - flat_start[n]
                    if flat_length >= ythresh:
                        # noise = np.random.rand(flat_length) * zthresh
                        noise = np.random.uniform(-1, 1, flat_length) * zthresh * 2
                        InteriorDomain[flat_start[n] : flat_stop[n], x] += noise

        return InteriorDomain

    def LTA_SC(
        self,
        InteriorDomain,
        OWloss,
        Qdg,
        DuneLoss,
        x_s,
        x_s_TS,
        x_t,
        x_t_TS,
        x_b_TS,
        s_sf_TS,
        InteriorWidth_Avg,
        SL,
        QowTS,
        QsfTS,
        t,
        h_b_TS,
    ):
        """Finds shoreline change for modeled year following Lorenzo-Trueba and Ashton
        (2014)"""

        # Find volume of shoreface/beach/dune sand deposited in island interior and
        # back-barrier
        Qow = OWloss / (
            self._BarrierLength
        )  # (dam^3/dam) Volume of sediment lost from shoreface/beach by overwash
        if Qow < 0:
            Qow = 0
        QowTS.append(Qow * 100)  # Save in m^3/m
        if DuneLoss < Qow:
            Qow = Qow - DuneLoss  # Account for dune contribution to overwash volume
        else:
            Qow = 0  # Excess DuneLoss assumed lost offshore

        if Qdg < 0:
            Qdg = 0

        # DefineParams
        d_sf = self._DShoreface
        h_b = np.average(InteriorDomain[InteriorDomain >= SL])
        x_b = x_s + InteriorWidth_Avg

        # Shoreface Flux
        s_sf = d_sf / (x_s - x_t)
        Qsf = (self._k_sf / 100) * (
            self._s_sf_eq - s_sf
        )  # Convert Ksf from m^3/m/yr to dam^3/dam/yr
        QsfTS.append(Qsf * 100)  # Save in m^3/m

        # Toe, Shoreline, and island base elevation changes
        x_t_dt = (4 * Qsf * (h_b + d_sf) / (d_sf * (2 * h_b + d_sf))) + (
            2 * self._RSLR[t] / s_sf
        )
        x_s_dt = 2 * (Qow + Qdg + self._Qat) / ((2 * h_b) + d_sf) - (
            4 * Qsf * (h_b + d_sf) / (((2 * h_b) + d_sf) ** 2)
        )  # Dune growth and alongshore transport added to LTA14 formulation

        # Record changes
        x_t = x_t + x_t_dt
        x_s = x_s + x_s_dt
        x_t_TS.append(x_t)
        x_s_TS.append(x_s)
        x_b_TS.append(x_b)
        s_sf_TS.append(s_sf)
        h_b_TS.append(h_b)

        return x_s, x_t, x_s_TS, x_t_TS, x_b_TS, s_sf_TS, QowTS, QsfTS, h_b_TS

    def CalcPC(
        self,
        ShrubDomainAll,
        PercentCoverTS,
        ShrubDomainDead,
        DeadPercentCoverTS,
        ShrubArea,
        t,
    ):
        """Calculates percent cover of shrub domain and shrub coverage area (dam^2)"""

        Allshrub_t = ShrubDomainAll.astype("int64")
        Deadshrub_t = ShrubDomainDead.astype("int64")
        ShrubPercentCover = self._PC.take(Allshrub_t)
        DeadPercentCover = self._PC.take(Deadshrub_t)

        # ShrubPercentCover[ShrubPercentCover==0] =
        #   DeadPercentCover[ShrubPercentCover==0]
        PercentCoverTS[t] = ShrubPercentCover
        DeadPercentCoverTS[t] = DeadPercentCover
        ShrubArea.append(np.count_nonzero(ShrubDomainAll))

        return (
            ShrubPercentCover,
            PercentCoverTS,
            ShrubArea,
            DeadPercentCover,
            DeadPercentCoverTS,
        )

    def migrate_dunes(
        self,
        shoreline_change_aggregate,
        cellular_shoreline_change,
        time_index,  # for model coupling
    ):
        if (
            shoreline_change_aggregate > 0
        ):  # Positive = prograde, add row(s) to front of interior domain
            for _ in range(cellular_shoreline_change):
                # New interior row added with elev. of previous dune field
                # (i.e. previous dune row now part of int.)
                self._InteriorDomain = np.vstack(
                    [
                        self._DuneDomain[time_index, :, -1] + self._BermEl,
                        self._InteriorDomain,
                    ]
                )
                if self._DuneParamStart == 0:
                    if self._seeded_RNG_on:
                        newDuneHeight = np.ones([self._BarrierLength]) * (
                            0.005
                            + (
                                -0.005
                                + (0.005 - (-0.005))
                                * self._RNG.random(self._BarrierLength)
                            )
                        )
                    else:
                        newDuneHeight = np.ones([self._BarrierLength]) * (
                            0.005
                            + (
                                -0.005
                                + (0.005 - (-0.005))
                                * np.random.rand(self._BarrierLength)
                            )
                        )
                else:
                    newDuneHeight = np.ones([self._BarrierLength]) * 0.01

                self._DuneDomain[time_index, :, :] = np.roll(
                    self._DuneDomain[time_index, :, :], 1, axis=1
                )
                self._DuneDomain[time_index, :, 0] = newDuneHeight

                # Update shrub domains too
                if self._Shrub_ON:
                    self._ShrubDomainFemale = np.concatenate(
                        (
                            np.zeros([1, self._BarrierLength]),
                            self._ShrubDomainFemale,
                        )
                    )
                    self._ShrubDomainMale = np.concatenate(
                        (np.zeros([1, self._BarrierLength]), self._ShrubDomainMale)
                    )
                    self._ShrubDomainDead = np.concatenate(
                        (np.zeros([1, self._BarrierLength]), self._ShrubDomainDead)
                    )
                    self._ShrubPercentCover = np.concatenate(
                        (
                            np.zeros([1, self._BarrierLength]),
                            self._ShrubPercentCover,
                        )
                    )
                    self._DeadPercentCover = np.concatenate(
                        (np.zeros([1, self._BarrierLength]), self._DeadPercentCover)
                    )
                    self._BurialDomain = np.concatenate(
                        (np.zeros([1, self._BarrierLength]), self._BurialDomain)
                    )

            # DomainWidth = DomainWidth + sc  # Update width of interior domain
            self._ShorelineChange = self._ShorelineChange + cellular_shoreline_change
            self._ShorelineChangeTS[time_index] = cellular_shoreline_change
            self._SCRagg[time_index] = (
                self._SCRagg[time_index] - cellular_shoreline_change
            )  # Reset, leaving residual

        elif (
            shoreline_change_aggregate < 0
        ):  # Negative = erode, remove front row(s) of interior domain
            for _ in range(cellular_shoreline_change):
                newDuneElev = self._InteriorDomain[
                    0, :
                ]  # First row of interior will now become new part of active dune field
                newDuneHeight = newDuneElev - self._BermEl
                newDuneHeight[newDuneHeight < self._DuneRestart] = self._DuneRestart
                self._InteriorDomain = np.delete(self._InteriorDomain, 0, axis=0)
                if np.shape(self._InteriorDomain)[0] <= 0:
                    self._drown_break = 1
                    break
                self._DuneDomain[time_index, :, :] = np.roll(
                    self._DuneDomain[time_index, :, :], -1, axis=1
                )
                self._DuneDomain[time_index, :, -1] = newDuneHeight

                # Update shrub domains too
                if self._Shrub_ON:
                    self._ShrubDomainFemale = np.delete(
                        self._ShrubDomainFemale, 0, axis=0
                    )
                    self._ShrubDomainMale = np.delete(self._ShrubDomainMale, 0, axis=0)
                    self._ShrubDomainDead = np.delete(self._ShrubDomainDead, 0, axis=0)
                    self._ShrubPercentCover = np.delete(
                        self._ShrubPercentCover, 0, axis=0
                    )
                    self._DeadPercentCover = np.delete(
                        self._DeadPercentCover, 0, axis=0
                    )
                    self._BurialDomain = np.delete(self._BurialDomain, 0, axis=0)

            # DomainWidth = DomainWidth - sc  # Update width of interior domain
            self._ShorelineChange = self._ShorelineChange - cellular_shoreline_change
            self._ShorelineChangeTS[
                time_index
            ] = (
                -cellular_shoreline_change
            )  # the floored number corresponding to cells moved
            self._SCRagg[time_index] = (
                self._SCRagg[time_index] + cellular_shoreline_change
            )  # Reset (make more positive), leaving residual

    def __init__(self, **kwds):
        self._RNG = kwds.pop("RNG")

        self._SL = 0  # Does not change (Lagrangian frame of reference)

        self._BarrierLength = kwds.pop("BarrierLength")
        self._BayDepth = kwds.pop("BayDepth")
        self._BermEl = kwds.pop("BermEl")
        self._beta = kwds.pop("beta")
        self._BurialLimit = kwds.pop("BurialLimit")
        self._C1 = kwds.pop("C1")
        self._C2 = kwds.pop("C2")
        self._Cbb_i = kwds.pop("Cbb_i")
        self._Cbb_r = kwds.pop("Cbb_r")
        self._Cx = kwds.pop("Cx")
        self._disp_mu = kwds.pop("disp_mu")
        self._disp_sigma = kwds.pop("disp_sigma")
        self._Dmaxel = kwds.pop("Dmaxel")
        self._InitialDomainWidth = kwds.pop("DomainWidth")
        self._DShoreface = kwds.pop("DShoreface")
        self._Dshrub = kwds.pop("Dshrub")
        Dstart = kwds.pop("Dstart")
        self._DuneDomain = kwds.pop("DuneDomain")
        self._DuneRestart = kwds.pop("DuneRestart")
        self._DuneWidth = kwds.pop("DuneWidth")
        self._Female = kwds.pop("Female")
        self._GermRate = kwds.pop("GermRate")
        self._growthparam = kwds.pop("growthparam")
        self._HdDiffu = kwds.pop("HdDiffu")
        self._InteriorDomain = kwds.pop("InteriorDomain")
        self._k_sf = kwds.pop("k_sf")
        self._Ki = kwds.pop("Ki")
        self._Kr = kwds.pop("Kr")
        LShoreface = kwds.pop("LShoreface")
        self._mm = kwds.pop("mm")
        self._MHW = kwds.pop("MHW")
        self._nn = kwds.pop("nn")
        self._OWss_i = kwds.pop("OWss_i")
        self._OWss_r = kwds.pop("OWss_r")
        self._PC = kwds.pop("PC")
        self._Qat = kwds.pop("Qat")
        self._Qs_bb_min = kwds.pop("Qs_bb_min")
        self._Qs_min = kwds.pop("Qs_min")
        self._Qshrub_max = kwds.pop("Qshrub_max")
        self._Rin_i = kwds.pop("Rin_i")
        self._Rin_r = kwds.pop("Rin_r")
        self._RSLR = kwds.pop("RSLR")
        self._s_sf_eq = kwds.pop("s_sf_eq")
        self._SalineLimit = kwds.pop("SalineLimit")
        self._Seedmax = kwds.pop("Seedmax")
        self._Seedmin = kwds.pop("Seedmin")
        self._Shrub_ON = kwds.pop("Shrub_ON")
        self._ShrubEl_max = kwds.pop("ShrubEl_max")
        self._ShrubEl_min = kwds.pop("ShrubEl_min")
        self._StormSeries = kwds.pop("StormSeries")
        self._StormStart = kwds.pop("StormStart")
        self._threshold_in = kwds.pop("threshold_in")
        self._TimeFruit = kwds.pop("TimeFruit")
        self._TMAX = kwds.pop("TMAX")
        self._UprootLimit = kwds.pop("UprootLimit")
        self._SprayDist = kwds.pop("SprayDist")
        self._MaxUpSlope = kwds.pop("MaxUpSlope")
        self._DuneParamStart = kwds.pop("DuneParamStart")
        self._GrowthParamStart = kwds.pop("GrowthParamStart")
        self._RSLR_Constant = kwds.pop("RSLR_Constant")
        self._RSLR_const = kwds.pop("RSLR_const")
        self._MaxShrubHeight = kwds.pop("MaxShrubHeight")
        self._x_t = kwds.pop("x_t")
        self._rmin = kwds.pop("rmin")
        self._rmax = kwds.pop("rmax")
        self._seeded_RNG_on = kwds.pop(
            "SeededRNG"
        )  # if True, use seeded random number generator

        if len(kwds) > 0:
            raise ValueError(
                "unrecognized keywords ({})".format(", ".join(kwds.keys()))
            )

        # ### Initialize Shrubs
        self._PercentCoverTS = [None] * self._TMAX
        self._PercentCoverTS[0] = np.zeros(
            [self._InitialDomainWidth, self._BarrierLength]
        )
        self._DeadPercentCoverTS = [None] * self._TMAX
        self._DeadPercentCoverTS[0] = np.zeros(
            [self._InitialDomainWidth, self._BarrierLength]
        )
        self._ShrubFemaleTS = [None] * self._TMAX
        self._ShrubFemaleTS[0] = np.zeros(
            [self._InitialDomainWidth, self._BarrierLength]
        )
        self._ShrubMaleTS = [None] * self._TMAX
        self._ShrubMaleTS[0] = np.zeros([self._InitialDomainWidth, self._BarrierLength])
        self._ShrubDeadTS = [None] * self._TMAX
        self._ShrubDeadTS[0] = np.zeros([self._InitialDomainWidth, self._BarrierLength])
        self._ShrubDomainFemale = np.zeros(
            [self._InitialDomainWidth, self._BarrierLength]
        )
        self._ShrubDomainMale = np.zeros(
            [self._InitialDomainWidth, self._BarrierLength]
        )
        self._ShrubDomainDead = np.zeros(
            [self._InitialDomainWidth, self._BarrierLength]
        )
        self._ShrubPercentCover = self._PercentCoverTS[0]
        self._DeadPercentCover = self._DeadPercentCoverTS[0]
        self._BurialDomain = np.zeros([self._InitialDomainWidth, self._BarrierLength])
        self._ShrubArea = [0]
        self._MaxAvgSlope = self._BermEl / 10

        # ### Initialize other variables
        self._DomainTS = [
            None
        ] * self._TMAX  # Stores the elevation domain for each timestep
        self._DomainTS[0] = self._InteriorDomain
        self._ShorelineChangeTS = np.zeros([self._TMAX])
        self._ShorelineChange = 0  # (dam) Total amount of shoreline change
        self._StormCount = [0]
        self._InundationCount = 0
        self._RunUpCount = 0
        self._QowTS = [0]  # (m^3/m)
        self._x_s = self._x_t + LShoreface  # (dam) Start location of shoreline
        self._x_t_TS = [self._x_t]  # (dam) Shoreface toe locations for each time step
        self._x_s_TS = [self._x_s]  # (dam) Shoreline locations for each time step
        # (dam) average height of barrier for each time step
        self._h_b_TS = [
            (np.average(self._InteriorDomain[self._InteriorDomain >= self._SL]))
        ]
        self._s_sf_TS = [(self._DShoreface / LShoreface)]
        # Counter variable for shoreline change that is less than 1 cell size per year
        self._SCRagg = np.zeros([self._TMAX])
        self._SCRagg[0] = (self._x_s % 1) * -1  # initialize with erosion
        self._Hd_AverageTS = [Dstart]
        self._QsfTS = [0]  # (m^3/m)
        self._Hd_Loss_TS = np.zeros(
            [self._TMAX, self._BarrierLength]
        )  # Dune height loss from vertical storm erosion

        # variables changed/added in Version 2.0
        # self._x_t = 0  # (dam) Start location of shoreface toe -- added as an
        # input variable
        # use FindWidths to calculate the average interior width for setting initial
        # back barrier shoreline
        _, _, InteriorWidth_Avg = self.FindWidths(self._InteriorDomain, self._SL)
        # self._InteriorWidth_AvgTS = [self._InitialDomainWidth]
        self._InteriorWidth_AvgTS = [InteriorWidth_Avg]
        # self._x_b_TS = [
        #     (self._x_s + self._InitialDomainWidth)
        # ]  # (dam) Bay shoreline locations for each time step
        self._x_b_TS = [
            (self._x_s + InteriorWidth_Avg)
        ]  # (dam) Bay shoreline locations for each time step
        self._drown_break = 0
        self._dune_migration_on = True  # allow dunes to migrate
        self._interior_noise_on = False  # add noise to flat parts of interior domain
        self._bay_routing_width = 10  # Width for bay section of overwash routing domain
        self._PreStorm_InteriorDomain = self._InteriorDomain

        self._time_index = 1

    @classmethod
    def from_path(cls, path_to_folder, prefix="barrier3d", fmt="yaml"):
        return cls(**load_inputs(path_to_folder, prefix=prefix, fmt=fmt))

    @classmethod
    def from_xlsx(cls, path_to_xlsx, prefix="barrier3d"):
        return cls(**load_inputs(path_to_xlsx, prefix=prefix, fmt="xlsx"))

    @classmethod
    def from_yaml(cls, path_to_yaml, prefix="barrier3d"):
        return cls(**load_inputs(path_to_yaml, prefix=prefix, fmt="yaml"))

    def update(self):
        # ###########################################
        # ### increase sea level and start new time step
        # ###########################################

        # ### RSLR
        self._InteriorDomain, self._DuneDomain = self.SeaLevel(
            self._InteriorDomain, self._DuneDomain, self._time_index
        )

        # ### Find DomainWidth and InteriorWidth
        DomainWidth, InteriorWidth, InteriorWidth_Avg = self.FindWidths(
            self._InteriorDomain, self._SL
        )

        # ### Grow Dunes
        self._DuneDomain, self._Dmax, Qdg = self.DuneGrowth(
            self._DuneDomain, self._time_index
        )

        DuneDomainCrest = self._DuneDomain[self._time_index, :, :].max(
            axis=1
        )  # Maximum height of each row in DuneDomain
        DuneDomainCrest[DuneDomainCrest < self._DuneRestart] = self._DuneRestart

        self._Hd_AverageTS.append(
            np.mean(DuneDomainCrest)
        )  # Store average pre-storm dune-height for time step

        # ###########################################
        # ### Shrubs

        if self._Shrub_ON:
            (
                self._ShrubDomainAll,
                self._ShrubDomainFemale,
                self._ShrubDomainMale,
                self._BurialDomain,
                self._ShrubDomainDead,
            ) = self.Shrubs(
                self._InteriorDomain,
                DuneDomainCrest,
                self._ShrubDomainFemale,
                self._ShrubDomainMale,
                self._ShrubPercentCover,
                self._DeadPercentCover,
                self._BurialDomain,
                InteriorWidth_Avg,
                DomainWidth,
                self._ShrubDomainDead,
            )

        # ###########################################
        # ### Storms

        OWloss = 0
        DuneLoss = 0
        numstorm = 0

        self._PreStorm_InteriorDomain = copy.deepcopy(self._InteriorDomain)

        if self._time_index >= self._StormStart:
            # Select number of storms for this time step from normal distribution
            TSloc = np.argwhere(self._StormSeries[:, 0] == self._time_index)
            numstorm = int(len(TSloc))  # analysis:ignore

            if numstorm > 0:
                start = TSloc[0, 0]
                stop = TSloc[-1, 0] + 1
                Rhigh = self._StormSeries[start:stop, 1]
                Rlow = self._StormSeries[start:stop, 2]
                dur = np.array(self._StormSeries[start:stop, 4], dtype="int")

                # ### Individual Storm Impacts
                for n in range(numstorm):  # Loop through each individual storm
                    # ###########################################
                    # ### Dune Erosion

                    DuneChange = np.zeros(
                        [self._BarrierLength, self._DuneWidth]
                    )  # Vector storing dune height change for this storm

                    # Find overwashed dunes and gaps
                    Dow = [
                        index
                        for index, value in enumerate(DuneDomainCrest + self._BermEl)
                        if value < Rhigh[n]
                    ]
                    gaps = self.DuneGaps(
                        DuneDomainCrest, Dow, self._BermEl, Rhigh[n]
                    )  # Finds location and Rexcess of continuous gaps in dune ridge

                    for d in range(len(Dow)):  # Loop through each overwashed dune cell
                        for w in range(self._DuneWidth):
                            # Calculate dune elevation loss
                            # Rhigh relative to pre-storm dune elevation
                            Rnorm = Rhigh[n] / (
                                self._DuneDomain[self._time_index, Dow[d], w]
                                + self._BermEl
                            )
                            # Amount of dune crest elevation change normalized by
                            # pre-storm dune elevation
                            Dloss = Rnorm / (self._C1 + (Rnorm * (Rnorm - self._C2)))
                            # (i.e. a percent change), from Goldstein and Moore (2016)

                            # Set new dune height
                            InitDElev = (
                                self._DuneDomain[self._time_index, Dow[d], w]
                                + self._BermEl
                            )
                            NewDElev = InitDElev * (
                                1 - Dloss
                            )  # Calculate new dune elevation from storm lowering
                            if NewDElev < self._BermEl:
                                NewDElev = self._BermEl
                            self._DuneDomain[self._time_index, Dow[d], w] = (
                                NewDElev - self._BermEl
                            )  # Convert elevation to height above berm

                            DuneChange[Dow[d], w] = InitDElev - NewDElev

                            # If dune is lowered to ~ zero, allow for chance of
                            # regrowth by raising dune height to 5 cm
                            if (
                                self._DuneDomain[self._time_index, Dow[d], w]
                                < self._DuneRestart
                            ):
                                if self._DuneRestart < self._Dmax:
                                    self._DuneDomain[
                                        self._time_index, Dow[d], w
                                    ] = self._DuneRestart
                                else:
                                    self._DuneDomain[
                                        self._time_index, Dow[d], w
                                    ] = (
                                        self._Dmax
                                    )  # Restart height can't be greater than Dmax

                    # Dune Height Diffusion
                    self._DuneDomain = self.DiffuseDunes(
                        self._DuneDomain, self._time_index
                    )
                    self._DuneDomain[self._DuneDomain < self._SL] = self._DuneRestart

                    DuneLoss = np.sum(DuneChange) / self._BarrierLength
                    Hd_TSloss = (
                        DuneChange.max(axis=1) / dur[n]
                    )  # Average height of dune loss for each substep during storm

                    self._Hd_Loss_TS[self._time_index, :] = self._Hd_Loss_TS[
                        self._time_index, :
                    ] + DuneChange.max(axis=1)

                    # ###########################################
                    # ### Overwash

                    Iow = 0  # Count of dune gaps in inundation regime
                    Dunes_prestorm = DuneDomainCrest
                    for q in range(len(gaps)):
                        start = gaps[q][0]
                        stop = gaps[q][1]
                        gapwidth = stop - start + 1
                        meandune = (
                            sum(Dunes_prestorm[start : stop + 1]) / gapwidth
                        ) + self._BermEl  # Average elevation of dune gap

                        # Determine number of gaps in inundation regime
                        if Rlow[n] > meandune:
                            Iow += 1

                    # Determine Sediment And Water Routing Rules
                    # If greater than threshold % of dune gaps are inunundation regime,
                    # use inun. regime routing
                    if len(gaps) > 0 and Iow / len(gaps) >= self._threshold_in:
                        inundation = 1
                        substep = self._OWss_i
                        self._InundationCount += 1
                    else:
                        inundation = 0
                        substep = self._OWss_r
                        self._RunUpCount += 1

                    # Set Domain
                    duration = dur[n] * substep
                    width = (
                        np.shape(self._InteriorDomain)[0] + 1 + self._bay_routing_width
                    )  # (dam) Add one for Dunes and 25 for bay
                    Elevation = np.zeros([duration, width, self._BarrierLength])
                    Dunes = Dunes_prestorm + self._BermEl
                    Bay = (
                        np.ones([self._bay_routing_width, self._BarrierLength])
                        * -self._BayDepth
                    )
                    Elevation[0, :, :] = np.vstack([Dunes, self._InteriorDomain, Bay])

                    # Initialize Memory Storage Arrays
                    Discharge = np.zeros([duration, width, self._BarrierLength])
                    SedFluxIn = np.zeros([duration, width, self._BarrierLength])
                    SedFluxOut = np.zeros([duration, width, self._BarrierLength])

                    # (dam^3/t) Infiltration Rate, volume of overwash flow lost per m
                    # cross-shore per time
                    Rin = 0

                    # Set Water at Dune Crest
                    for q in range(len(gaps)):
                        start = gaps[q][0]
                        stop = gaps[q][1]
                        Rexcess = gaps[q][2]  # (m)

                        # Calculate discharge through each dune cell
                        Vdune = math.sqrt(2 * 9.8 * (Rexcess * 10)) / 10  # (dam/s)
                        Qdune = Vdune * Rexcess * 3600  # (dam^3/hr)

                        # Set discharge at dune gap
                        Discharge[:, 0, start:stop] = Qdune

                        if inundation == 1:  # Inundation regime
                            Rin = self._Rin_i

                            # # Find average slope of interior
                            # AvgSlope = self._BermEl / InteriorWidth_Avg
                            #
                            # # Enforce max average interior slope
                            # AvgSlope = min(self._MaxAvgSlope, AvgSlope)

                            # Representative average slope of interior (made static -
                            # represent. of 200-m wide barrier)
                            AvgSlope = self._BermEl / 20

                            C = self._Cx * AvgSlope  # Momentum constant

                        else:  # Run-up regime
                            Rin = self._Rin_r

                    # ### Run Flow Routing Algorithm
                    for TS in range(duration):
                        ShrubDomainWidth = np.shape(self._ShrubDomainFemale)[0]
                        DeadDomainWidth = np.shape(self._ShrubDomainDead)[0]

                        if TS > 0:
                            Elevation[TS, 1:, :] = Elevation[
                                TS - 1, 1:, :
                            ]  # Begin timestep with elevation from end of last
                            Elevation[TS, 0, :] = Dunes - (
                                Hd_TSloss / substep * TS
                            )  # Reduce dune in height linearly over course of storm

                        for d in range(width - 1):
                            # Reduce discharge across row via infiltration
                            if d > 0:
                                Discharge[TS, d, :][Discharge[TS, d, :] > 0] -= Rin
                            Discharge[TS, d, :][Discharge[TS, d, :] < 0] = 0

                            for i in range(self._BarrierLength):
                                if Discharge[TS, d, i] > 0:
                                    Q0 = Discharge[TS, d, i]

                                    # ### Calculate Slopes
                                    if i > 0:
                                        S1 = (
                                            Elevation[TS, d, i]
                                            - Elevation[TS, d + 1, i - 1]
                                        ) / (math.sqrt(2))
                                        S1 = np.nan_to_num(S1)
                                    else:
                                        S1 = 0

                                    S2 = Elevation[TS, d, i] - Elevation[TS, d + 1, i]
                                    S2 = np.nan_to_num(S2)

                                    if i < (self._BarrierLength - 1):
                                        S3 = (
                                            Elevation[TS, d, i]
                                            - Elevation[TS, d + 1, i + 1]
                                        ) / (math.sqrt(2))
                                        S3 = np.nan_to_num(S3)
                                    else:
                                        S3 = 0

                                    # ### Calculate Discharge To Downflow Neighbors
                                    # One or more slopes positive
                                    if S1 > 0 or S2 > 0 or S3 > 0:
                                        if S1 < 0:
                                            S1 = 0
                                        if S2 < 0:
                                            S2 = 0
                                        if S3 < 0:
                                            S3 = 0

                                        Q1 = (
                                            Q0
                                            * S1**self._nn
                                            / (
                                                S1**self._nn
                                                + S2**self._nn
                                                + S3**self._nn
                                            )
                                        )
                                        Q2 = (
                                            Q0
                                            * S2**self._nn
                                            / (
                                                S1**self._nn
                                                + S2**self._nn
                                                + S3**self._nn
                                            )
                                        )
                                        Q3 = (
                                            Q0
                                            * S3**self._nn
                                            / (
                                                S1**self._nn
                                                + S2**self._nn
                                                + S3**self._nn
                                            )
                                        )

                                        Q1 = np.nan_to_num(Q1)
                                        Q2 = np.nan_to_num(Q2)
                                        Q3 = np.nan_to_num(Q3)

                                    # No slopes positive, one or more equal to zero
                                    elif S1 == 0 or S2 == 0 or S3 == 0:
                                        pos = 0
                                        if S1 == 0:
                                            pos += 1
                                        if S2 == 0:
                                            pos += 1
                                        if S3 == 0:
                                            pos += 1

                                        Qx = Q0 / pos
                                        Qx = np.nan_to_num(Qx)

                                        if S1 == 0 and i > 0:
                                            Q1 = Qx
                                        else:
                                            Q1 = 0
                                        if S2 == 0:
                                            Q2 = Qx
                                        else:
                                            Q2 = 0
                                        if S3 == 0 and i < (self._BarrierLength - 1):
                                            Q3 = Qx
                                        else:
                                            Q3 = 0

                                    # All slopes negative
                                    else:
                                        Q1 = (
                                            Q0
                                            * abs(S1) ** (-self._nn)
                                            / (
                                                abs(S1) ** (-self._nn)
                                                + abs(S2) ** (-self._nn)
                                                + abs(S3) ** (-self._nn)
                                            )
                                        )
                                        Q2 = (
                                            Q0
                                            * abs(S2) ** (-self._nn)
                                            / (
                                                abs(S1) ** (-self._nn)
                                                + abs(S2) ** (-self._nn)
                                                + abs(S3) ** (-self._nn)
                                            )
                                        )
                                        Q3 = (
                                            Q0
                                            * abs(S3) ** (-self._nn)
                                            / (
                                                abs(S1) ** (-self._nn)
                                                + abs(S2) ** (-self._nn)
                                                + abs(S3) ** (-self._nn)
                                            )
                                        )

                                        Q1 = np.nan_to_num(Q1)
                                        Q2 = np.nan_to_num(Q2)
                                        Q3 = np.nan_to_num(Q3)

                                        # MaxUpSlope = 0.25  # dam

                                        if S1 > self._MaxUpSlope:
                                            Q1 = 0
                                        else:
                                            Q1 = Q1 * (1 - (abs(S1) / self._MaxUpSlope))

                                        if S2 > self._MaxUpSlope:
                                            Q2 = 0
                                        else:
                                            Q2 = Q2 * (1 - (abs(S2) / self._MaxUpSlope))

                                        if S3 > self._MaxUpSlope:
                                            Q3 = 0
                                        else:
                                            Q3 = Q3 * (1 - (abs(S3) / self._MaxUpSlope))

                                    # Reduce Overwash Through Shrub Cells and Save
                                    # Discharge
                                    if self._Shrub_ON:
                                        # Cell 1
                                        if i > 0:
                                            if (
                                                d < ShrubDomainWidth
                                                and self._ShrubPercentCover[d, i - 1]
                                                > 0
                                            ):
                                                Q1 = Q1 * (
                                                    (1 - self._Qshrub_max)
                                                    * self._ShrubPercentCover[d, i - 1]
                                                )
                                            elif (
                                                d < DeadDomainWidth
                                                and self._DeadPercentCover[d, i - 1] > 0
                                            ):
                                                # Dead shrubs block 2/3 of what living
                                                # shrubs block
                                                Q1 = Q1 * (
                                                    (1 - self._Qshrub_max * 0.66)
                                                    * self._DeadPercentCover[d, i - 1]
                                                )
                                            Discharge[TS, d + 1, i - 1] = (
                                                Discharge[TS, d + 1, i - 1] + Q1
                                            )

                                        # Cell 2
                                        # (ShrubDomainWidth - 1)?
                                        if (
                                            d < ShrubDomainWidth
                                            and self._ShrubPercentCover[d, i] > 0
                                        ):
                                            Q2 = Q2 * (
                                                (1 - self._Qshrub_max)
                                                * self._ShrubPercentCover[d, i]
                                            )
                                        elif (
                                            d < DeadDomainWidth
                                            and self._DeadPercentCover[d, i] > 0
                                        ):
                                            Q2 = Q2 * (
                                                (1 - self._Qshrub_max * 0.66)
                                                * self._DeadPercentCover[d, i]
                                            )
                                        Discharge[TS, d + 1, i] = (
                                            Discharge[TS, d + 1, i] + Q2
                                        )

                                        # Cell 3
                                        if i < (self._BarrierLength - 1):
                                            if (
                                                d < ShrubDomainWidth
                                                and self._ShrubPercentCover[d, i + 1]
                                                > 0
                                            ):
                                                Q3 = Q3 * (
                                                    (1 - self._Qshrub_max)
                                                    * self._ShrubPercentCover[d, i + 1]
                                                )
                                            elif (
                                                d < DeadDomainWidth
                                                and self._DeadPercentCover[d, i + 1] > 0
                                            ):
                                                Q3 = Q3 * (
                                                    (1 - self._Qshrub_max * 0.66)
                                                    * self._DeadPercentCover[d, i + 1]
                                                )
                                            Discharge[TS, d + 1, i + 1] = (
                                                Discharge[TS, d + 1, i + 1] + Q3
                                            )
                                    else:
                                        # Cell 1
                                        if i > 0:
                                            Discharge[TS, d + 1, i - 1] = (
                                                Discharge[TS, d + 1, i - 1] + Q1
                                            )

                                        # Cell 2
                                        Discharge[TS, d + 1, i] = (
                                            Discharge[TS, d + 1, i] + Q2
                                        )

                                        # Cell 3
                                        if i < (self._BarrierLength - 1):
                                            Discharge[TS, d + 1, i + 1] = (
                                                Discharge[TS, d + 1, i + 1] + Q3
                                            )

                                    # ### Calculate Sed Movement
                                    fluxLimit = self._Dmax

                                    # Run-up Regime
                                    if inundation == 0:
                                        if Q1 > self._Qs_min and S1 >= 0:
                                            Qs1 = self._Kr * Q1
                                            if Qs1 > fluxLimit:
                                                Qs1 = fluxLimit
                                        else:
                                            Qs1 = 0

                                        if Q2 > self._Qs_min and S2 >= 0:
                                            Qs2 = self._Kr * Q2
                                            if Qs2 > fluxLimit:
                                                Qs2 = fluxLimit
                                        else:
                                            Qs2 = 0

                                        if Q3 > self._Qs_min and S3 >= 0:
                                            Qs3 = self._Kr * Q3
                                            if Qs3 > fluxLimit:
                                                Qs3 = fluxLimit
                                        else:
                                            Qs3 = 0

                                    # Inundation Regime - Murray and Paola (1994, 1997)
                                    # Rule 3 with flux limiter
                                    else:
                                        if Q1 > self._Qs_min:
                                            Qs1 = self._Ki * (Q1 * (S1 + C)) ** self._mm
                                            if Qs1 < 0:
                                                Qs1 = 0
                                            elif Qs1 > fluxLimit:
                                                Qs1 = fluxLimit
                                        else:
                                            Qs1 = 0

                                        if Q2 > self._Qs_min:
                                            Qs2 = self._Ki * (Q2 * (S2 + C)) ** self._mm
                                            if Qs2 < 0:
                                                Qs2 = 0
                                            elif Qs2 > fluxLimit:
                                                Qs2 = fluxLimit
                                        else:
                                            Qs2 = 0

                                        if Q3 > self._Qs_min:
                                            Qs3 = self._Ki * (Q3 * (S3 + C)) ** self._mm
                                            if Qs3 < 0:
                                                Qs3 = 0
                                            elif Qs3 > fluxLimit:
                                                Qs3 = fluxLimit
                                        else:
                                            Qs3 = 0

                                    Qs1 = np.nan_to_num(Qs1)
                                    Qs2 = np.nan_to_num(Qs2)
                                    Qs3 = np.nan_to_num(Qs3)

                                    ### Calculate Net Erosion/Accretion
                                    if Elevation[TS, d, i] > self._SL or any(
                                        z > self._SL
                                        for z in Elevation[TS, d + 1 : d + 10, i]
                                    ):
                                        # If cell is subaerial, elevation change is
                                        # determined by difference between
                                        # flux in vs. flux out
                                        if i > 0:
                                            SedFluxIn[TS, d + 1, i - 1] += Qs1

                                        SedFluxIn[TS, d + 1, i] += Qs2

                                        if i < (self._BarrierLength - 1):
                                            SedFluxIn[TS, d + 1, i + 1] += Qs3

                                        Qs_out = Qs1 + Qs2 + Qs3
                                        SedFluxOut[TS, d, i] = Qs_out

                                    else:
                                        # If cell is subaqeous, exponentially decay
                                        # dep. of remaining sed across bay
                                        if inundation == 0:
                                            Cbb = self._Cbb_r
                                        else:
                                            Cbb = self._Cbb_i

                                        Qs0 = SedFluxIn[TS, d, i] * Cbb

                                        Qs1 = Qs0 * Q1 / (Q1 + Q2 + Q3)
                                        Qs2 = Qs0 * Q2 / (Q1 + Q2 + Q3)
                                        Qs3 = Qs0 * Q3 / (Q1 + Q2 + Q3)

                                        Qs1 = np.nan_to_num(Qs1)
                                        Qs2 = np.nan_to_num(Qs2)
                                        Qs3 = np.nan_to_num(Qs3)

                                        if Qs1 < self._Qs_bb_min:
                                            Qs1 = 0
                                        elif Qs1 > fluxLimit:
                                            Qs1 = fluxLimit
                                        if Qs2 < self._Qs_bb_min:
                                            Qs2 = 0
                                        elif Qs2 > fluxLimit:
                                            Qs2 = fluxLimit
                                        if Qs3 < self._Qs_bb_min:
                                            Qs3 = 0
                                        elif Qs3 > fluxLimit:
                                            Qs3 = fluxLimit

                                        if i > 0:
                                            SedFluxIn[TS, d + 1, i - 1] += Qs1

                                        SedFluxIn[TS, d + 1, i] += Qs2

                                        if i < (self._BarrierLength - 1):
                                            SedFluxIn[TS, d + 1, i + 1] += Qs3

                                        Qs_out = Qs1 + Qs2 + Qs3
                                        SedFluxOut[TS, d, i] = Qs_out

                                    if self._Shrub_ON:
                                        # ### Saline Flooding
                                        (
                                            self._ShrubDomainFemale,
                                            self._ShrubDomainMale,
                                            self._ShrubDomainDead,
                                        ) = self.SalineFlooding(
                                            ShrubDomainWidth,
                                            self._ShrubDomainAll,
                                            self._ShrubDomainFemale,
                                            self._ShrubDomainMale,
                                            self._ShrubDomainDead,
                                            d,
                                            i,
                                            Q0,
                                        )

                        # ### Update Elevation After Every Storm Hour
                        if inundation == 1:
                            ElevationChange = (
                                SedFluxIn[TS, :, :] - SedFluxOut[TS, :, :]
                            ) / substep
                        else:
                            ElevationChange = (
                                SedFluxIn[TS, :, :] - SedFluxOut[TS, :, :]
                            ) / substep
                        Elevation[TS, :, :] = Elevation[TS, :, :] + ElevationChange

                        # Calculate and save volume of sediment deposited on/behind the
                        # island for every hour
                        # (all four methods below should equal the same!)
                        # OWloss = OWloss + np.sum(ElevationChange[1:,:])
                        # OWloss = OWloss + (
                        #   np.sum(SedFluxIn[TS,1:,:]) - np.sum(SedFluxOut[TS,1:,:])
                        # ) / substep
                        # OWloss = OWloss + np.sum(SedFluxIn[TS,1,:]) / substep
                        OWloss = OWloss + np.sum(SedFluxOut[TS, 0, :]) / substep

                        if self._Shrub_ON:
                            # Update amount of burial/erosion for each shrub
                            self._BurialDomain = self.UpdateBurial(
                                self._BurialDomain,
                                ElevationChange,
                                ShrubDomainWidth,
                                self._ShrubDomainAll,
                            )

                    # ### Update Interior Domain After Every Storm
                    InteriorUpdate = Elevation[-1, 1:, :]

                    # Remove all rows of bay without any deposition from the domain
                    check = 1
                    while check == 1:
                        if all(x <= -self._BayDepth for x in InteriorUpdate[-1, :]):
                            InteriorUpdate = np.delete(InteriorUpdate, (-1), axis=0)
                        else:
                            check = 0

                    # Update interior domain
                    self._InteriorDomain = InteriorUpdate

                    # Update Domain widths
                    DomainWidth = np.shape(self._InteriorDomain)[0]
                    if self._Shrub_ON:
                        (
                            self._ShrubDomainFemale,
                            self._ShrubDomainMale,
                            self._ShrubDomainAll,
                            self._ShrubPercentCover,
                            self._BurialDomain,
                            self._ShrubDomainDead,
                            self._DeadPercentCover,
                        ) = self.UpdateShrubDomains(
                            DomainWidth,
                            ShrubDomainWidth,
                            self._ShrubDomainFemale,
                            self._ShrubDomainMale,
                            self._ShrubDomainAll,
                            self._ShrubPercentCover,
                            self._BurialDomain,
                            self._ShrubDomainDead,
                            self._DeadPercentCover,
                        )

        # Record storm data
        self._StormCount.append(numstorm)

        if self._interior_noise_on:  # add noise to flat parts of interior domain
            if self._time_index > self._StormStart:
                self._InteriorDomain = self.InteriorNoise(
                    self._InteriorDomain, InteriorWidth
                )

        # ###########################################
        # ### Ocean Shoreline Change

        # Calculate shoreline/shoreface position change following Lorenzo-Trueba
        # and Ashton (2014)
        (
            self._x_s,
            self._x_t,
            self._x_s_TS,
            self._x_t_TS,
            self._x_b_TS,
            self._s_sf_TS,
            self._QowTS,
            self._QsfTS,
            self._h_b_TS,
        ) = self.LTA_SC(
            self._InteriorDomain,
            OWloss,
            Qdg,
            DuneLoss,
            self._x_s,
            self._x_s_TS,
            self._x_t,
            self._x_t_TS,
            self._x_b_TS,
            self._s_sf_TS,
            InteriorWidth_Avg,
            self._SL,
            self._QowTS,
            self._QsfTS,
            self._time_index,
            self._h_b_TS,
        )

    def update_dune_domain(self):
        # ###########################################
        # ### update dune domain (erode/prograde) based on on shoreline change
        # ###########################################

        self._drown_break = 0

        if self._dune_migration_on:
            # (dam) Change rate of ocean-fronting shoreline:
            # (+) = progradation, (-) = erosion
            SCR = (self._x_s_TS[-1] - self._x_s_TS[-2]) * -1
            # Account for any residual shoreline change (less than cell size) from
            # previous time step
            self._SCRagg[self._time_index] = self._SCRagg[self._time_index - 1] + SCR

            # if shoreline moved to the next cell, prograde or erode dune line
            if abs(self._SCRagg[self._time_index]) >= 1:
                sc = math.floor(abs(self._SCRagg[self._time_index]))

                self.migrate_dunes(
                    self._SCRagg[self._time_index],
                    sc,
                    self._time_index,
                )

            else:
                self._ShorelineChangeTS[
                    self._time_index
                ] = 0  # redundant because the new time array is zeros, but keep

        # ### Check for drowning
        if self._drown_break == 1:
            self._TMAX = self._time_index - 1
            print(
                "Barrier has WIDTH DROWNED at t = {time} years".format(
                    time=self._time_index
                )
            )
            return  # exit program
        elif all(j <= self._SL for j in self._InteriorDomain[0, :]):
            self._TMAX = self._time_index - 1
            print(
                "Barrier has HEIGHT DROWNED at t = {time} years".format(
                    time=self._time_index
                )
            )
            self._drown_break = 1
            return  # exit program

        # ### Recalculate and save DomainWidth and InteriorWidth
        DomainWidth, InteriorWidth, InteriorWidth_Avg = self.FindWidths(
            self._InteriorDomain, self._SL
        )
        self._InteriorWidth_AvgTS.append(InteriorWidth_Avg)

        # replace value of x_b_TS with new InteriorWidth_Avg
        self._x_b_TS[-1] = self._x_s_TS[-1] + self._InteriorWidth_AvgTS[-1]

        # ###########################################
        # ### Save domains
        # ###########################################

        self._DomainTS[self._time_index] = self._InteriorDomain

        if self._Shrub_ON:
            zero = np.zeros([DomainWidth, self._BarrierLength])
            # Correctly saves domain to list only if array of zeros are added to domain
            # first, no clue why - 4Mar20
            self._ShrubFemaleTS[self._time_index] = self._ShrubDomainFemale + zero
            self._ShrubMaleTS[self._time_index] = self._ShrubDomainMale + zero
            self._ShrubDeadTS[self._time_index] = self._ShrubDomainDead + zero

            # Calculate Percent Cover and Area
            self._ShrubDomainAll = self._ShrubDomainMale + self._ShrubDomainFemale
            (
                self._ShrubPercentCover,
                self._PercentCoverTS,
                self._ShrubArea,
                self._DeadPercentCover,
                self._DeadPercentCoverTS,
            ) = self.CalcPC(
                self._ShrubDomainAll,
                self._PercentCoverTS,
                self._ShrubDomainDead,
                self._DeadPercentCoverTS,
                self._ShrubArea,
                self._time_index,
            )

        self._time_index += 1

    @property
    def time_index(self):
        return self._time_index

    @property
    def QowTS(self):
        return self._QowTS

    @property
    def QsfTS(self):
        return self._QsfTS

    @property
    def x_t_TS(self):
        return self._x_t_TS

    @property
    def x_s_TS(self):
        return self._x_s_TS

    @x_s_TS.setter
    def x_s_TS(self, value):
        self._x_s_TS = value

    @property
    def x_s(self):
        return self._x_s

    @x_s.setter
    def x_s(self, value):
        self._x_s = value

    @property
    def x_t(self):
        return self._x_t

    @x_t.setter
    def x_t(self, value):
        self._x_t = value

    @property
    def x_b_TS(self):
        return self._x_b_TS

    @property
    def growthparam(self):
        return self._growthparam

    @growthparam.setter
    def growthparam(self, value):
        self._growthparam = value

    @property
    def DuneDomain(self):
        return self._DuneDomain

    @DuneDomain.setter
    def DuneDomain(self, value):
        self._DuneDomain = value

    @property
    def InteriorDomain(self):
        return self._InteriorDomain

    @InteriorDomain.setter
    def InteriorDomain(self, value):
        self._InteriorDomain = value

    @property
    def DomainTS(self):
        return self._DomainTS

    @DomainTS.setter
    def DomainTS(self, value):
        self._DomainTS = value

    @property
    def BermEl(self):
        return self._BermEl

    @property
    def rmin(self):
        return self._rmin

    @property
    def rmax(self):
        return self._rmax

    @property
    def Dmax(self):
        return self._Dmax

    @property
    def StormSeries(self):
        return self._StormSeries

    @StormSeries.setter
    def StormSeries(self, value):
        self._StormSeries = value

    @property
    def StormStart(self):
        return self._StormStart

    @StormStart.setter
    def StormStart(self, value):
        self._StormStart = value

    @property
    def InteriorWidth_AvgTS(self):
        return self._InteriorWidth_AvgTS

    @InteriorWidth_AvgTS.setter
    def InteriorWidth_AvgTS(self, value):
        self._InteriorWidth_AvgTS = value

    @property
    def drown_break(self):
        return self._drown_break

    @property
    def s_sf_TS(self):
        return self._s_sf_TS

    @s_sf_TS.setter
    def s_sf_TS(self, value):
        self._s_sf_TS = value

    @property
    def h_b_TS(self):
        return self._h_b_TS

    @h_b_TS.setter
    def h_b_TS(self, value):
        self._h_b_TS = value

    @property
    def DShoreface(self):
        return self._DShoreface

    @DShoreface.setter
    def DShoreface(self, value):
        self._DShoreface = value

    @property
    def dune_migration_on(self):
        return self._dune_migration_on

    @dune_migration_on.setter
    def dune_migration_on(self, value):
        self._dune_migration_on = value

    @property
    def TMAX(self):
        return self._TMAX

    @property
    def bay_routing_width(self):
        return self._bay_routing_width

    @property
    def BarrierLength(self):
        return self._BarrierLength

    @property
    def ShrubDomainFemale(self):
        return self._ShrubDomainFemale

    @property
    def ShrubDomainMale(self):
        return self._ShrubDomainMale

    @property
    def ShrubDomainAll(self):
        return self._ShrubDomainAll

    @property
    def ShrubDomainDead(self):
        return self._ShrubDomainDead

    @property
    def ShrubPercentCover(self):
        return self._ShrubPercentCover

    @property
    def DeadPercentCover(self):
        return self._DeadPercentCover

    @property
    def BurialDomain(self):
        return self._BurialDomain

    @property
    def ShorelineChange(self):
        return self._ShorelineChange

    @property
    def ShorelineChangeTS(self):
        return self._ShorelineChangeTS

    @property
    def RSLR(self):
        return self._RSLR

    @RSLR.setter
    def RSLR(self, value):
        self._RSLR = value

    @property
    def SL(self):
        return self._SL

    @property
    def Hd_AverageTS(self):
        return self._Hd_AverageTS

    @property
    def Dmaxel(self):
        return self._Dmaxel

    @property
    def DuneWidth(self):
        return self._DuneWidth

    @property
    def BayDepth(self):
        return self._BayDepth

    @property
    def DuneRestart(self):
        return self._DuneRestart

    @property
    def SCRagg(self):
        return self._SCRagg

    @property
    def PreStorm_InteriorDomain(self):
        return self._PreStorm_InteriorDomain



from ._version import __version__
from .barrier3d import Barrier3d
from .bmi import Barrier3dBmi
from .configuration import Barrier3dConfiguration
from .load_input import load_inputs

__all__ = [
    "__version__",
    "Barrier3d",
    "Barrier3dBmi",
    "Barrier3dConfiguration",
    "load_inputs",
]



# Static length (alongshore) of island segment
BarrierLength = 500.0  # [m]
# Depth of bay behind island segment
BayDepth = 3.0  # [m]
# Static elevation of berm; berm elevation + dune height = dune elevation (NOTE: if
# changed, need new MSSM and storms)
BermEl = 1.9  # [m NAVD88]
# Maximum percentage of height that a shrub can be buried up to before dying
BurialLimit = 0.75  # [m]
# Empirical dune erosion parameter
C1 = 8.8  # [m]
# Empirical dune erosion parameter
C2 = 4.6  # [m]
# Coefficient for exponential decay of sediment load entering back-barrier bay in
# inundation regime
Cbb_i = 0.85
# Coefficient for exponential decay of sediment load entering back-barrier bay in run-up
# regime
Cbb_r = 0.7
# Multiplier with the average slope of the interior for constant C in inundation
# transport rule
Cx = 10.0
# Height of shoreface
DShoreface = 10.0  # [m]
# Maximum elevation of dunes
Dmaxel = 3.4  # [m NAVD88]
# Minimum elevation of fronting dune for shrub growth
Dshrub = 2.75  # [m]
# Initial height of dune domain above berm elevation
Dstart = 0.5  # [m]
# Dune height will come from external file
DuneParamStart = True
# Restart height for dunes lowered to essentially zero
DuneRestart = 0.075  # [m]
# Width (cross-shore) of island dune field; for illustration purposes only
DuneWidth = 20.0  # [m]
# Percentage of shrubs that are female
Female = 0.5
# Germination rate
GermRate = 0.6
# Dune growth parameters will come from external file
GrowthParamStart = True
# Dune diffusion parameter (i.e. max height offset between adjacent dune cells)
HdDiffu = 0.75  # [m]
# Sediment flux constant, inundation regime
Ki = 7.5e-06
# Sediment flux constant, run-up regime
Kr = 7.5e-05
# Length of shoreface
LShoreface = 500.0  # [m]
# Elevation of Mean High Water (NOTE: if changed, need new storm time series)
MHW = 0.46  # [m NAVD88]
# Maximum shrub height
MaxShrubHeight = 5.3  # [m]
# Maximum slope water can flow upward
MaxUpSlope = 0.25  # [m / m]
# Overwash substep
OWss_i = 2
# Overwash substep
OWss_r = 1
# Minimum sediment flux in back-barrier bay (below which sediment won't flux)
Qs_bb_min = 1.0  # [m^3 / hr]
# Minimum discharge needed for sediment transport
Qs_min = 1.0  # [m^3 / hr]
# Maximum percentage of overwash reduction through a shrub cell with full percent cover
Qshrub_max = 0.15
# Relative sea-level rise rate will be constant, otherwise logistic growth function used
# for time series
RSLR_Constant = True
# Relative sea-level rise rate
RSLR_const = 0.004  # [m / y]
# Rate of shoreline retreat attributed to alongshore transport; (-) = erosion, (+) =
# accretion
Rat = 0.0  # [m / y]
# Inundation regime infiltration rate (volume of overwash flow lost per m cross-shore
# per time step)
Rin_i = 0.1
# Run-up regime infiltration rate (volume of overwash flow lost per m cross-shore per
# time step)
Rin_r = 2.0
# Dishcharge limit to determine shrub mortality via saline flooding
SalineLimit = 5.0  # [m^3 / hr]
# Use seeded random number generator for reproducibility
SeededRNG = True
# Seeds produced per shrub per year (fecundity)
Seedmax = 1000.0  # [1 / yr]
# Seeds produced per shrub per year (fecundity)
Seedmin = 100.0  # [1 / yr]
# Start location of shoreface toe [m]
ShorefaceToe = 0.0
# Elevation range for shrub growth, maximum bound
ShrubEl_max = 2.3  # [m NAVD88]
# Elevation range for shrub growth, minimum bound
ShrubEl_min = 1.2  # [m NAVD88]
# 1 = shrubs on in simulation, 0 = shrubs off
Shrub_ON = False
# Distance from ocean shoreline that shrubs can establish
SprayDist = 170.0  # [m]
# Time series of storms
StormSeries = []
# Year when storm can start occurring (NOTE: if changed, need new storm time series)
StormStart = 2  # [y]
# Duration of simulation
TMAX = 51  # [y]
# Age shrubs need to be before they start fruiting
TimeFruit = 5.0  # [y]
# Shrubs eroded beyond this limit killed
UprootLimit = -0.2  # [m]
# Beach slope for runup calculations
beta = 0.04
# For lognormal probability distribution of seed dispersal distance
disp_mu = -0.721891
# For lognormal probability distribution of seed dispersal distance
disp_sigma = 1.5
# File that contains initial dune height values [m]
dune_file = "barrier3d-default-dunes.npy"
# File that contains initial elevations in [m MHH]
elevation_file = "barrier3d-default-elevations.npy"
# File that contains initial growth parameters
growth_param_file = "barrier3d-default-growthparam.npy"
# Shoreface flux rate constant
k_sf = 5000.0  # [m^3 / m / y]
# Exponent constant for sediment transport
mm = 2.0
# Flow routing constant
nn = 0.5
# Maximum growth rate for logistic dune growth
rmax = 0.85
# Minimum growth rate for logistic dune growth
rmin = 0.35
# Equilibrium shoreface slope
s_sf_eq = 0.02
# File that contains storm data
storm_file = "barrier3d-default-storms.npy"
# Threshold to determine if in inundation regime
threshold_in = 0.25  # [m^3 / hr]



"""
These functions create, return, and if specified save time series of
annual storm parameters, initial dune height, and dune growth rates
for use as inputs in Barrier3D simulations.

References
----------
.. [1] Wahl, T., Plant, N. G., & Long, J. W. (2016). Probabilistic assessment of
   erosion and flooding risk in the northern Gulf of Mexico. Journal of Geophysical
   Research: Oceans, 121(5), 3029-3043.

Notes
-----

The storm time series can be created using one of three functions, all of
which require a list of storms generated using the multivariateSeaStorm.m
module, developed following the method of ...[1]:
1. yearly_storms - specify the mean and standard deviation
   from a normal distribution, which are used to select a random number
   of storms per year from the MSSM list
2. frequency_storms - specify the TWL and frequency of a
   return period storm, find closest match in MSSM list
3. shift_storm_intensity - shifts the TWL distribution created in 1) to
   the left or right to produce storms of different "intensity".

"""
import bisect
import pathlib
import random

import matplotlib.pyplot as plt
import numpy as np
from distfit import distfit

from ..load_input import _guess_format


def yearly_storms(
    datadir=".",
    storm_list_name="StormList_20k_VCR_Berm1pt9m_Slope0pt04.csv",  # can by .py or .csv
    mean_yearly_storms=8.3,
    SD_yearly_storms=5.9,
    MHW=0.46,  # m NAVD88
    StormStart=2,
    BermEl=1.9,  # m NAVD88, just used for plotting
    model_years=10000,
    bPlot=True,
    bSave=False,
    output_filename="StormList_10kyrs_VCR_Berm1pt9m_Slope0pt04",
):
    """
    Use a normal distribution (provided the mean and standard deviation)
    to select a random number of storms per year from a list of
    multivariate sea storms, created using the Wahl et al., 2016 method.
    """

    datadir = pathlib.Path(datadir)

    # convert to decameters
    MHW = MHW / 10
    BermEl = BermEl / 10 - MHW  # just for plotting

    # load list of storms (created using multivariateSeaStorm.m)
    fmt = _guess_format(datadir / storm_list_name)
    if fmt == "npy":
        StormList = np.load(datadir / storm_list_name, allow_pickle=True)
    elif fmt == "csv":
        StormList = np.loadtxt(
            datadir / storm_list_name, delimiter=",", encoding="utf-8-sig"
        )

    # pad with zeros until storms start
    StormSeries = np.zeros([StormStart, 5])

    for t in range(StormStart, model_years):
        # Calculate number of storms in year
        numstorm = round(np.random.normal(mean_yearly_storms, SD_yearly_storms))
        if numstorm < 0:
            numstorm = 0
        stormTS = np.zeros([numstorm, 5])

        # Select storms for year
        for n in range(numstorm):
            storm = random.randint(1, len(StormList) - 1)

            dur = StormList[storm, 1]  # Duration
            Rhigh = StormList[storm, 2]  # TWL
            period = StormList[storm, 4]  # Tp
            Rlow = StormList[storm, 6]  # Rlow

            stormTS[n, 0] = t
            stormTS[n, 1] = Rhigh / 10 - MHW  # make relative to MHW
            stormTS[n, 2] = Rlow / 10 - MHW
            stormTS[n, 3] = period
            stormTS[n, 4] = round(
                dur / 2
            )  # Divided by two assuming TWL only for only half of storm
            # (NOTE from KA: we could probably do better here, like assume that the
            # TWL follows a distribution; or create a time series as in the Wahl 2019
            # follow up paper for MSSM. Future work!)

        StormSeries = np.vstack([StormSeries, stormTS])

    if bPlot:
        # Plots
        Bin = np.linspace(-1, 4.6, 57)

        plt.figure()
        surgetidecalc = StormSeries[:, 2] * 10
        plt.hist(surgetidecalc, bins=Bin)
        plt.title("StormSeries Rlow")
        plt.xlabel("m MHW")

        plt.figure()
        twlcalc = StormSeries[:, 1] * 10
        plt.hist(twlcalc, bins=Bin)
        plt.title("StormSeries Rhigh")
        plt.xlabel("m MHW")

        plt.figure()
        fig = plt.gcf()
        fig.set_size_inches(16, 4)
        plt.plot(twlcalc)
        plt.plot(
            np.arange(0, len(twlcalc), 1),
            np.ones(len(twlcalc)) * BermEl * 10,
            "r--",
        )
        plt.xlabel("Storm")
        plt.ylabel("TWL (m MHW)")
        plt.legend(["TWL", "BermEl"])

        print("Max TWL (m MHW): ", max(StormSeries[:, 1]) * 10)
        print("Max Rexcess (m above berm): ", (max(StormSeries[:, 1]) - BermEl) * 10)
        print(
            "% Rhigh > BermEl: ", (twlcalc > (BermEl * 10)).sum() / len(twlcalc) * 100
        )
        print(
            "% Rlow  > BermEl: ",
            (surgetidecalc > (BermEl * 10)).sum() / len(surgetidecalc) * 100,
        )

    if bSave:
        np.save(datadir / output_filename, StormSeries)

    return StormSeries


def shift_storm_intensity(
    datadir=".",
    storm_list_name="StormList_20k_VCR_Berm1pt9m_Slope0pt04.csv",  # can by .py or .csv
    mean_yearly_storms=8.3,
    SD_yearly_storms=5.9,
    # shift the TWL distribution to change intensity, m NAVD88; [-0.15, 0.15] for
    # Reeves et al., 2021
    shift=-0.15,
    MHW=0.46,  # m NAVD88
    StormStart=2,
    BermEl=1.9,  # m NAVD88, just used for plotting
    model_years=1000,
    bPlot=True,
    bSave=False,
    output_filename="StormList_10kyrs_VCR_Berm1pt9m_Slope0pt04-lowIntensity",
):
    """
    Fit a beta distribution to the TWL time series and then shift the beta
    distribution to the left or right to simulate TWLs of higher or lower intensities
    """
    datadir = pathlib.Path(datadir)

    # convert to decameters
    MHW = MHW / 10
    BermEl = BermEl / 10 - MHW  # just for plotting

    # load list of storms (created using multivariateSeaStorm.m)
    fmt = _guess_format(datadir / storm_list_name)
    if fmt == "npy":
        StormList = np.load(datadir / storm_list_name, allow_pickle=True)
    elif fmt == "csv":
        StormList = np.loadtxt(datadir / storm_list_name, delimiter=",")

    # sort the storms based on TWL, from min to max (probably a more elegant way to
    # do this)
    dur = StormList[:, 1]  # Duration
    simTWL = StormList[:, 2]
    period = StormList[:, 4]  # Tp
    Rlow = StormList[:, 6]  # Rlow
    zip_storm_list = list(zip(simTWL, dur, period, Rlow))
    zip_storm_list.sort()
    simTWL_sorted = np.array([simTWL for (simTWL, dur, period, Rlow) in zip_storm_list])
    dur_sorted = np.array([dur for (simTWL, dur, period, Rlow) in zip_storm_list])
    period_sorted = np.array([period for (simTWL, dur, period, Rlow) in zip_storm_list])
    Rlow_sorted = np.array([Rlow for (simTWL, dur, period, Rlow) in zip_storm_list])

    # Fit Distribution - Note: this is typically beta for VCR TWLs, so we assume beta
    # here
    dist = distfit(distr="beta")
    fit = dist.fit_transform(simTWL_sorted)
    dist.plot()
    model_parameters = dist.model  # dictionary of beta distribution parameters
    loc = model_parameters["loc"]
    scale = model_parameters["scale"]
    [a, b] = model_parameters["arg"]

    # Bin Specifications
    BinStart = round(np.min(fit["histdata"][1]), 1)  # round to one decimal
    BinStop = round(np.max(fit["histdata"][1]), 1)
    BinWidth = 0.1  # m
    BinNum = int(((BinStop - BinStart) / BinWidth) + 2)
    Bin = np.linspace(BinStart, BinStop, BinNum)

    # Make storm series
    StormSeries = np.zeros([StormStart, 5])

    for t in range(StormStart, model_years):
        # Calculate number of storms in year
        numstorm = max(0, round(np.random.normal(mean_yearly_storms, SD_yearly_storms)))
        stormTS = np.zeros([numstorm, 5])

        # Select storms for year
        for n in range(numstorm):
            ST = (
                np.random.beta(a, b) * scale + loc + shift
            )  # the TWL from the beta distribution
            if ST < 0:
                ST = 0
            elif ST > simTWL_sorted[-1]:
                ST = simTWL_sorted[-1]
            STmin = Bin[bisect.bisect(Bin, ST) - 1]
            STmax = STmin + 0.1
            indexMin = bisect.bisect(simTWL_sorted, STmin)
            indexMax = bisect.bisect(simTWL_sorted, STmax) - 1
            if indexMin == 0:
                storm = 1
            elif indexMin >= indexMax:
                storm = indexMin
            else:
                storm = random.randint(indexMin, indexMax)

            dur = dur_sorted[storm]
            Rhigh = simTWL_sorted[storm]
            period = period_sorted[storm]
            Rlow = Rlow_sorted[storm]

            stormTS[n, 0] = t
            stormTS[n, 1] = Rhigh / 10 - MHW
            stormTS[n, 2] = Rlow / 10 - MHW
            stormTS[n, 3] = period
            stormTS[n, 4] = round(
                dur / 2
            )  # Divided by two assuming TWL only for only half of storm

        # Save
        StormSeries = np.vstack([StormSeries, stormTS])

    if bPlot:
        # Plots
        Bin = np.linspace(-1, 4.6, 57)

        plt.figure()
        surgetidecalc = StormSeries[:, 2] * 10
        plt.hist(surgetidecalc, bins=Bin)
        plt.title("StormSeries Rlow")
        plt.xlabel("m MHW")

        plt.figure()
        twlcalc = StormSeries[:, 1] * 10
        plt.hist(twlcalc, bins=Bin)
        plt.title("StormSeries Rhigh")
        plt.xlabel("m MHW")

        plt.figure()
        fig = plt.gcf()
        fig.set_size_inches(16, 4)
        plt.plot(twlcalc)
        plt.plot(
            np.arange(0, len(twlcalc), 1),
            np.ones(len(twlcalc)) * BermEl * 10,
            "r--",
        )
        plt.xlabel("Storm")
        plt.ylabel("TWL (m MHW)")
        plt.legend(["TWL", "BermEl"])

        print("Max TWL (m MHW): ", max(StormSeries[:, 1]) * 10)
        print("Max Rexcess (m above berm): ", (max(StormSeries[:, 1]) - BermEl) * 10)
        print(
            "% Rhigh > BermEl: ", (twlcalc > (BermEl * 10)).sum() / len(twlcalc) * 100
        )
        print(
            "% Rlow  > BermEl: ",
            (surgetidecalc > (BermEl * 10)).sum() / len(surgetidecalc) * 100,
        )

    if bSave:
        np.save(datadir / output_filename, StormSeries)

    return StormSeries


def frequency_storms(
    datadir,
    storm_list_name="StormList_20k_VCR_Berm1pt9m_Slope0pt04.csv",  # can by .py or .csv
    MHW=0.46,
    return_period=50,  # minimum of 1
    # in m above MHW (from NOAA Annual Exceedance Probability Curves)
    return_period_TWL=2.0,
    StormStart=2,
    BermEl=1.9,  # m NAVD88, just used for plotting
    model_years=1000,
    bPlot=True,
    bSave=False,
    output_filename="StormList_50yrRP_2mTWL_1kyrs_VCR_Berm1pt9m_Slope0pt04",
):
    """
    Select a storm from the list of multivariate sea storms -- created using
    the Wahl et al., 2016 method -- that matches the TWL specified for a
    given return period (in m above MHW), and returns the storm time
    series at the specified return period
    """

    datadir = pathlib.Path(datadir)

    # convert to decameters
    MHW = MHW / 10
    BermEl = BermEl / 10 - MHW  # just for plotting
    return_period_TWL = return_period_TWL / 10

    # load list of storms (created using multivariateSeaStorm.m)
    fmt = _guess_format(datadir / storm_list_name)
    if fmt == "npy":
        StormList = np.load(datadir / storm_list_name, allow_pickle=True)
    elif fmt == "csv":
        StormList = np.loadtxt(datadir / storm_list_name, delimiter=",")

    # find storm that has the closest TWL (in m above MHW) to the return period storm
    dur = np.round(
        StormList[:, 1] / 2
    )  # Duration, divided by two assuming TWL only for only half of storm
    Rhigh = StormList[:, 2] / 10 - MHW  # TWL in dam above MHW
    period = StormList[:, 4]  # Tp
    Rlow = StormList[:, 6] / 10 - MHW  # Rlow in dam relative to MHW

    [id, closest_TWL] = min(
        enumerate(Rhigh), key=lambda x: abs(x[1] - return_period_TWL)
    )

    # pad with zeros until storms start
    StormSeries = np.zeros([StormStart, 5])

    for t in range(StormStart, model_years):
        # only allow for one storm per year
        numstorm = 1
        stormTS = np.zeros([numstorm, 5])

        # Select storms for year
        if t % return_period == 0:
            stormTS[0, 0] = t
            stormTS[0, 1] = Rhigh[id]
            stormTS[0, 2] = Rlow[id]
            stormTS[0, 3] = period[id]
            stormTS[0, 4] = dur[id]

        StormSeries = np.vstack([StormSeries, stormTS])

    if bPlot:
        # Plots
        Bin = np.linspace(-1, 4.6, 57)

        plt.figure()
        surgetidecalc = StormSeries[:, 2] * 10
        plt.hist(surgetidecalc, bins=Bin)
        plt.title("StormSeries Rlow")
        plt.xlabel("m MHW")

        plt.figure()
        twlcalc = StormSeries[:, 1] * 10
        plt.hist(twlcalc, bins=Bin)
        plt.title("StormSeries Rhigh")
        plt.xlabel("m MHW")

        plt.figure()
        fig = plt.gcf()
        fig.set_size_inches(16, 4)
        plt.plot(twlcalc)
        plt.plot(
            np.arange(0, len(twlcalc), 1),
            np.ones(len(twlcalc)) * BermEl * 10,
            "r--",
        )
        plt.xlabel("Storm")
        plt.ylabel("TWL (m MHW)")
        plt.legend(["TWL", "BermEl"])

        print("Max TWL (m MHW): ", max(StormSeries[:, 1]) * 10)
        print("Max Rexcess (m above berm): ", (max(StormSeries[:, 1]) - BermEl) * 10)
        print(
            "% Rhigh > BermEl: ", (twlcalc > (BermEl * 10)).sum() / len(twlcalc) * 100
        )
        print(
            "% Rlow  > BermEl: ",
            (surgetidecalc > (BermEl * 10)).sum() / len(surgetidecalc) * 100,
        )

    if bSave:
        np.save(datadir / output_filename, StormSeries)

    return StormSeries


def gen_dune_height_start(datadir, name, Dstart=0.5, ny=1000):
    """Generate dune height start"""
    datadir = pathlib.Path(datadir)

    # convert to decameters
    Dstart = Dstart / 10

    DuneStart = np.ones([ny]) * (
        Dstart + (-0.01 + (0.01 - (-0.01)) * np.random.rand(ny))
    )

    return np.save(datadir / name, DuneStart)


def gen_alongshore_variable_rmin_rmax(datadir, name, rmin=0.35, rmax=0.85, ny=1000):
    """Generate along-shore varying rmin & rmax"""
    datadir = pathlib.Path(datadir)

    growthparam = rmin + (rmax - rmin) * np.random.rand(ny)

    return np.save(datadir / name, growthparam)



"""
    Example run script for the BMI version of Barrier3D (Version 2.0)

    NOTE to users:
        - if using Barrier3D for the first time, remember to $ pip install -e .
"""

import time

from barrier3d import Barrier3dBmi
from barrier3d.tools import plot as B3Dfunc

# create an instance of the new BMI class, which is the model
barrier3d = Barrier3dBmi()

# specify data directory with initial conditions
datadir = "tests/test_params/"
input_file = "barrier3d-default-parameters.yaml"
barrier3d.initialize(datadir + input_file)

# increase time step
Time = time.time()
for time_step in range(1, barrier3d._model._TMAX):
    barrier3d.update()  # update the model by a time step

    # Print time step to screen
    print("\r", "Time Step: ", time_step, end="")

SimDuration = time.time() - Time
print()
print("Elapsed Time: ", SimDuration, "sec")  # Print elapsed time of simulation

# Plot 1: Dune Height Over Time (input in decameter)
B3Dfunc.plot_dune_height(barrier3d._model._DuneDomain, barrier3d._model._Dmax)
B3Dfunc.plot_shrub_percent_cover_tmax(
    barrier3d._model._PercentCoverTS,
    barrier3d._model._TMAX,
    barrier3d._model._DeadPercentCoverTS,
)
B3Dfunc.plot_ElevTMAX(
    barrier3d._model._TMAX,
    barrier3d._model._DuneDomain,
    barrier3d._model._DomainTS,
    barrier3d._model._BermEl,
    barrier3d._model._Shrub_ON,
    barrier3d._model._PercentCoverTS,
    barrier3d._model._DeadPercentCoverTS,
    barrier3d._model._DuneWidth,
)



#! /usr/bin/env python
import os
import pathlib
from functools import partial

import click
import numpy as np
import pandas as pd
import pkg_resources

from . import __version__
from .barrier3d import Barrier3d
from .configuration import Barrier3dConfiguration

out = partial(click.secho, bold=True, err=True)
err = partial(click.secho, fg="red", err=True)


def csv_contents(infile):
    datadir = pathlib.Path(pkg_resources.resource_filename("barrier3d", "data"))
    with open(datadir / f"barrier3d-default-{infile}.csv") as fp:
        return fp.read()


INFILES = {
    "parameters": Barrier3dConfiguration().to_yaml(),
    "elevations": csv_contents("elevations"),
    "dunes": csv_contents("dunes"),
    "growthparam": csv_contents("growthparam"),
    "storms": csv_contents("storms"),
}


class Barrier3dOutputWriter:
    def __init__(self, bmi, output_interval=1, filepath="output.csv"):
        self._bmi = bmi
        self._output_interval = output_interval
        self._steps = 0

        with open(filepath, mode="w") as fp:
            print(f"# version: {__version__}", file=fp)
            print(
                ",".join(
                    [
                        "# Time step [-]",
                        "Overwash flux [m^3/m]",
                        "Shoreface flux [m^3/m]",
                        "Change in shoreface toe position [dam]",
                        "Change in shoreline position [dam]",
                        "Back-barrier shoreline position [dam]",
                    ]
                ),
                file=fp,
            )

    def _save_output(self):
        data = pd.DataFrame(
            {
                "time_stamp": self._steps,
                "q_ow": self._bmi.QowTS[-1],
                "q_sf": self._bmi.QsfTS[-1],
                "dx_toe_dt": np.diff(self._bmi.x_t_TS[-2:]),
                "dx_sf_dt": np.diff(self._bmi.x_s_TS[-2:]),
                "x_b": self._bmi.x_b_TS[-1],
            },
            index=(0,),
        )
        data.to_csv("output.csv", mode="a", index=False, sep=",", header=False)

    def update(self, n_steps):
        self._steps += n_steps
        if self._steps % self._output_interval == 0:
            self._save_output()


@click.group(chain=True)
@click.version_option()
@click.option(
    "--cd",
    default=".",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, readable=True),
    help="chage to directory, then execute",
)
def barrier3d(cd) -> None:
    """Barrier3D: A spatially explicit exploratory model of barrier island evolution
    in three dimensions

    \b
    Examples:

      Create a folder with example input files,

        $ mkdir example && b3d setup

      Run a simulation using the examples input files,

        $ cd example && b3d run

      Commands can also be chained together. The following will setup
      a simulation, run it, and then plot elevations.

        $ mkdir example && b3d setup run plot elevation
    """
    os.chdir(cd)


@barrier3d.command()
@click.option("-v", "--verbose", is_flag=True, help="Emit status messages to stderr.")
@click.option("--dry-run", is_flag=True, help="Do not actually run the model")
def run(dry_run: bool, verbose: bool) -> None:
    """Run a simulation."""
    run_dir = pathlib.Path.cwd()
    # config_file = run_dir / "barrier3d.yaml"

    message = []
    # if not config_file.is_file():
    #     message.append(f"missing Barrier3D configuration file: {config_file}")
    if (run_dir / "output.csv").exists():
        message.append(
            "Barrier3D output file already exists: {}".format(run_dir / "output.csv")
        )
    if message:
        err(os.linesep.join(message))
        raise click.Abort(os.linesep.join(message))

    barrier3d = Barrier3d.from_path(run_dir, fmt="yaml")
    output = Barrier3dOutputWriter(barrier3d, filepath=run_dir / "output.csv")

    n_steps = barrier3d._TMAX - 1

    if dry_run:
        out(str(run_dir))
    elif n_steps == 0:
        out("Nothing to do (years == 0). 😴")
    else:
        with click.progressbar(
            range(n_steps),
            label=" ".join(["🚀", str(run_dir)]),
            item_show_func=lambda step: "step {} of {}".format(
                int(0 if step is None else step), n_steps
            ),
        ) as bar:
            for _ in bar:
                barrier3d.update()
                output.update(1)

        out("💥 Finished! 💥")


@barrier3d.command()
@click.argument("infile", type=click.Choice(sorted(INFILES)))
def show(infile: str) -> None:
    """Show example input files."""
    print(_contents_of_input_file(infile))


@barrier3d.command()
def setup() -> None:
    """Setup a folder of input files for a simulation."""
    files = {
        "parameters": pathlib.Path("barrier3d-default-parameters.yaml"),
        "elevations": pathlib.Path("barrier3d-default-elevations.csv"),
        "dunes": pathlib.Path("barrier3d-default-dunes.csv"),
        "growthparam": pathlib.Path("barrier3d-default-growthparam.csv"),
        "storms": pathlib.Path("barrier3d-default-storms.csv"),
    }

    existing_files = [name for name in files.values() if name.exists()]
    if existing_files:
        for name in existing_files:
            err(
                f"{name}: File exists. Either remove and then rerun or setup in a "
                "different folder",
            )
    else:
        for infile, fname in files.items():
            with open(fname, "w") as fp:
                print(_contents_of_input_file(infile), file=fp)

    if existing_files:
        raise click.Abort()


def _contents_of_input_file(infile: str) -> str:
    if infile not in INFILES:
        raise ValueError(
            "unknown input file type ({infile} not one of {infiles})".format(
                infile=infile, infiles=", ".join(sorted(INFILES))
            )
        )

    return INFILES[infile]


@barrier3d.command()
@click.argument(
    "value",
    type=click.Choice(["time_step", "q_ow", "q_sf", "dx_toe_dt", "dx_sf_dt", "x_bb"]),
)
def plot(value: str) -> None:
    """Plot output from a simulation."""
    import matplotlib.pyplot as plt

    filepath = pathlib.Path("output.csv")
    out(str(filepath.resolve()))

    data = pd.read_csv(
        filepath,
        names=("time_step", "q_ow", "q_sf", "dx_toe_dt", "dx_sf_dt", "x_bb"),
        comment="#",
    )
    plt.plot(data["time_step"], data[value])

    plt.show()



import difflib
import os
import pathlib
import shutil
import tempfile

import nox

PROJECT = "barrier3d"
ROOT = pathlib.Path(__file__).parent


nox.options.sessions = ["lint", "test", "test-bmi", "test-cli", "test-notebooks"]


@nox.session
def test(session: nox.Session) -> None:
    """Run the tests."""
    session.install("-r", "requirements-testing.txt")
    session.install(".")

    args = [
        "-n",
        "auto",
        "--cov",
        PROJECT,
        "-vvv",
    ] + session.posargs

    if "CI" in os.environ:
        args.append(f"--cov-report=xml:{ROOT.absolute()!s}/coverage.xml")
    session.run("pytest", *args)

    if "CI" not in os.environ:
        session.run("coverage", "report", "--ignore-errors", "--show-missing")


@nox.session(name="test-versions")
def test_versions(session: nox.Session):
    """Test with the previous version of barrier3d."""
    session.install("-r", "requirements-testing.txt")
    session.install("-r", "version1_local_copy/requirements.txt")
    session.install(".")

    session.run("pytest", "-vvv", "-m", "slow")


@nox.session(name="test-cli")
def test_cli(session: nox.Session):
    """Test the b3d commandline interface."""
    session.install(".")

    session.run("b3d", "--help")
    session.run("b3d", "--version")
    session.run("b3d", "plot", "--help")
    session.run("b3d", "run", "--help")
    session.run("b3d", "setup", "--help")
    session.run("b3d", "show", "--help")

    for infile in ["dunes", "elevations", "growthparam", "parameters", "storms"]:
        out = session.run("b3d", "show", infile, external=True, silent=True)
        if not out.strip():
            session.error(f"The command `b3d show {infile}` produced no output")

    expected = {
        "barrier3d-default-dunes.csv",
        "barrier3d-default-elevations.csv",
        "barrier3d-default-growthparam.csv",
        "barrier3d-default-parameters.yaml",
        "barrier3d-default-storms.csv",
    }

    with tempfile.TemporaryDirectory() as tmpdirname:
        with session.chdir(tmpdirname):
            session.run("b3d", "setup")
            actual = {str(p) for p in pathlib.Path(".").iterdir()}

    if actual != expected:
        session.log(
            os.linesep.join(
                ["diff actual expected"]
                + list(
                    difflib.unified_diff(
                        sorted(actual),
                        sorted(expected),
                        fromfile="actual",
                        tofile="expected",
                    )
                )
            )
        )
        session.error("The command `b3d setup` generated an unexpected set of files")


@nox.session(name="test-bmi", venv_backend="mamba")
def test_bmi(session: nox.Session) -> None:
    """Run the tests."""
    session.conda_install("bmi-tester")
    session.conda_install("--file", "requirements.txt")
    session.install(".", "--no-deps")

    session.run(
        "bmi-test",
        "--config-file=tests/test_bmi/barrier3d.yaml",
        "--root-dir=tests/test_bmi",
        "-vvv",
        "barrier3d.bmi:Barrier3dBmi",
    )


@nox.session(name="test-notebooks")
def test_notebooks(session: nox.Session) -> None:
    """Run the notebooks."""
    args = [
        "pytest",
        "notebooks",
        "--nbmake",
        "--nbmake-kernel=python3",
        "--nbmake-timeout=3000",
        "-n",
        "auto",
        "-vvv",
    ] + session.posargs

    session.install("-r", "requirements-testing.txt")
    session.install("nbmake")
    session.install("-r", "notebooks/requirements.txt")
    session.install(".")

    session.run(*args)


@nox.session
def lint(session: nox.Session) -> None:
    """Look for lint."""
    session.install("pre-commit")
    session.run("pre-commit", "run", "--all-files")


@nox.session
def build(session: nox.Session) -> None:
    """Build sdist and wheel dists."""
    session.install("pip")
    session.install("build")
    session.run("python", "--version")
    session.run("pip", "--version")
    session.run("python", "-m", "build", "--outdir", "./build/wheelhouse")


@nox.session
def release(session):
    """Tag, build and publish a new release to PyPI."""
    session.install("zest.releaser[recommended]")
    session.run("fullrelease")


@nox.session(name="publish-testpypi")
def publish_testpypi(session):
    """Publish wheelhouse/* to TestPyPI."""
    session.install("twine")
    session.run("twine", "check", "build/wheelhouse/*")
    session.run(
        "twine",
        "upload",
        "--skip-existing",
        "--repository-url",
        "https://test.pypi.org/legacy/",
        "build/wheelhouse/*.tar.gz",
    )


@nox.session(name="publish-pypi")
def publish_pypi(session):
    """Publish wheelhouse/* to PyPI."""
    session.install("twine")
    session.run("twine", "check", "build/wheelhouse/*")
    session.run(
        "twine",
        "upload",
        "--skip-existing",
        "build/wheelhouse/*.tar.gz",
    )


@nox.session(python=False)
def clean(session):
    """Remove all .venv's, build files and caches in the directory."""
    root_folders = (
        [
            "barrier3d.egg-info",
            ".pytest_cache",
            ".venv",
            "build",
            "build/wheelhouse",
        ]
        if not session.posargs
        else []
    )

    with session.chdir(ROOT):
        for folder in root_folders:
            session.log(f"rm -r {folder}")
            shutil.rmtree(folder, ignore_errors=True)

    for folder in _args_to_folders(session.posargs):
        with session.chdir(folder):
            for pattern in ["*.py[co]", "__pycache__"]:
                session.log(f"rm {pattern}")
                _clean_rglob(pattern)


@nox.session(python=False, name="clean-checkpoints")
def clean_checkpoints(session):
    """Remove jupyter notebook checkpoint files."""
    for folder in _args_to_folders(session.posargs):
        with session.chdir(folder):
            _clean_rglob("*-checkpoint.ipynb")
            _clean_rglob(".ipynb_checkpoints")


def _args_to_folders(args):
    return [ROOT] if not args else [pathlib.Path(f) for f in args]


def _clean_rglob(pattern):
    nox_dir = pathlib.Path(".nox")

    for p in pathlib.Path(".").rglob(pattern):
        if nox_dir in p.parents:
            continue
        if p.is_dir():
            p.rmdir()
        else:
            p.unlink()



import importlib

import pytest
from numpy.testing import assert_array_almost_equal

from barrier3d import Barrier3d, load_inputs
from barrier3d.load_input import as_cwd


@pytest.mark.skip("missing file")
def test_parameters_to_expected(datadir):
    with as_cwd(datadir):
        mod = importlib.import_module("barrier3d-parameters-expected")
        expected = {k: v for k, v in mod.__dict__.items() if not k.startswith("_")}
        actual = load_inputs(datadir, prefix="barrier3d-default", fmt="yaml")

    for key in actual:
        assert key in expected
        try:
            assert actual[key] == expected[key], key
        except ValueError:
            assert_array_almost_equal(actual[key], expected[key], err_msg=key)
    # assert actual == expected


def test_barrier3d_configuration(datadir):
    expected = load_inputs(datadir, prefix="barrier3d-default", fmt="py")
    actual = load_inputs(datadir, prefix="barrier3d-default", fmt="yaml")

    assert set(actual.keys()) - set(expected.keys()) == set(), "found extra keys"
    assert set(expected.keys()) - set(actual.keys()) == set(), "missing keys"

    actual = dict(actual.items())

    assert_array_almost_equal(
        actual.pop("InteriorDomain"), expected.pop("InteriorDomain")
    )
    assert_array_almost_equal(actual.pop("PC"), expected.pop("PC"))
    assert_array_almost_equal(actual.pop("growthparam"), expected.pop("growthparam"))
    assert_array_almost_equal(actual.pop("DuneDomain"), expected.pop("DuneDomain"))
    assert_array_almost_equal(actual.pop("StormSeries"), expected.pop("StormSeries"))

    actual.pop("RNG")
    expected.pop("RNG")

    for key in expected:
        assert actual[key] == expected[key], f"mismatch in values for {key}"


def test_bad_input_format(datadir):
    with pytest.raises(ValueError):
        load_inputs(datadir, prefix="barrier3d-default", fmt="xlsx")


@pytest.mark.parametrize("fmt", ["yaml", "py"])
def test_missing_file(datadir, fmt):
    with pytest.raises(FileNotFoundError, match=f"missing-prefix-parameters.{fmt}"):
        load_inputs(datadir, prefix="missing-prefix", fmt=fmt)


def test_missing_folder(datadir):
    with pytest.raises(FileNotFoundError, match="missing-folder"):
        load_inputs("missing-folder", fmt="yaml")


@pytest.mark.parametrize("fmt", ["yaml", "py"])
def test_barrier3d_init(datadir, fmt):
    params = load_inputs(datadir, prefix="barrier3d-default", fmt=fmt)
    barrier3d = Barrier3d(**params)
    assert isinstance(barrier3d, Barrier3d)


@pytest.mark.parametrize("fmt", ["yaml", "py"])
def test_barrier3d_from_path(datadir, fmt):
    barrier3d = Barrier3d.from_path(datadir, prefix="barrier3d-default", fmt=fmt)
    assert isinstance(barrier3d, Barrier3d)


@pytest.mark.parametrize("fmt", ["yaml", "py"])
def test_barrier3d_update(datadir, fmt):
    barrier3d = Barrier3d.from_path(datadir, prefix="barrier3d-default", fmt=fmt)
    barrier3d.update()
